import gym

import numpy as np

from functools import reduce
from operator import mul


def make_grid_from_ranges(xi: list) -> list:
    """Create a grid of points from a list of ranges.

    Args:
        xi: List of ranges.

    Returns:
        Grid of points (the product of all points in ranges).

    Example:

        >>> import numpy as np
        >>> from gym_socks.utils.grid import make_grid_from_ranges
        >>> grid = make_grid_from_ranges([np.linspace(-1, 1, 5), np.linspace(-1, 1, 5)])

    """

    len_xi = len(xi)

    dtype = np.result_type(*xi)
    result = np.empty([len(a) for a in xi] + [len_xi], dtype=dtype)

    for i, x in enumerate(np.ix_(*xi)):
        result[..., i] = x

    return result.reshape(-1, len_xi)


def make_grid_from_space(sample_space: gym.spaces.Box, resolution: int) -> list:
    """Create a grid from a bounded sample space.

    Args:
        sample_space: Bounded sample space.
        resolution: The gid resolution, specified either as an integer for all
            dimensions or as a list of integers, one per dimension.

    Returns:
        Grid of points (the product of all points in ranges).

    Example:

        >>> from gym.spaces import Box
        >>> import numpy as np
        >>> from gym_socks.utils.grid import make_grid_from_space
        >>> sample_space = Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        >>> grid = make_grid_from_space(sample_space, [3, 2])

    """

    assert sample_space.is_bounded(), "space must be bounded."

    low = sample_space.low
    high = sample_space.high

    if np.isscalar(resolution):
        xi = np.linspace(low, high, resolution, axis=-1)
        grid = make_grid_from_ranges(xi)

    else:
        assert (
            np.shape(resolution) == sample_space.shape
        ), "resolution.shape doesn't match sample_space.shape"
        xi = []
        for i, value in enumerate(resolution):
            xi.append(np.linspace(low[i], high[i], value))

        grid = make_grid_from_ranges(xi)

    return grid


def grid_size_from_ranges(xi: list) -> int:
    """Returns the size of a grid sample based on ranges.

    Args:
        xi: List of ranges.

    Returns:
        The product of the lengths of the ranges.

    """

    return reduce(mul, map(len, xi), 1)


def grid_size_from_space(sample_space: gym.spaces.Box, resolution: int) -> int:
    """Returns the size of a grid sample based on sample space and resolution.

    Args:
        sample_space: The sample space.
        resolution: The gid resolution, specified either as an integer for all
            dimensions or as a list of integers, one per dimension.

    Returns:
        The product of the lengths of the ranges.

    """

    assert sample_space.is_bounded(), "space must be bounded."

    low = sample_space.low
    high = sample_space.high

    if np.isscalar(resolution):
        resolution = np.full(shape=sample_space.shape, fill_value=resolution)

    else:
        assert (
            np.shape(resolution) == sample_space.shape
        ), "resolution.shape doesn't match sample_space.shape"

    return reduce(mul, resolution, 1)



import unittest

import gym

import gym_socks.utils

import numpy as np

from gym_socks.utils.grid import make_grid_from_ranges
from gym_socks.utils.grid import make_grid_from_space
from gym_socks.utils.grid import grid_size_from_ranges
from gym_socks.utils.grid import grid_size_from_space


class TestGrid(unittest.TestCase):
    def test_grid_from_ranges(cls):
        """Should generate proper grid from ranges."""

        xi = [np.linspace(-1, 1, 3), np.linspace(-1, 1, 3)]
        result = make_grid_from_ranges(xi)

        groundTruth = [
            [-1.0, -1.0],
            [-1.0, 0.0],
            [-1.0, 1.0],
            [0.0, -1.0],
            [0.0, 0.0],
            [0.0, 1.0],
            [1.0, -1.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]

        cls.assertTrue(np.array_equiv(result, groundTruth))

    def test_grid_from_space(cls):
        """Should generate proper grid from space."""

        sample_space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        groundTruth = [
            [-1.0, -1.0],
            [-1.0, 0.0],
            [-1.0, 1.0],
            [0.0, -1.0],
            [0.0, 0.0],
            [0.0, 1.0],
            [1.0, -1.0],
            [1.0, 0.0],
            [1.0, 1.0],
        ]

        result = make_grid_from_space(sample_space=sample_space, resolution=3)
        cls.assertTrue(np.array_equiv(result, groundTruth))

        result = make_grid_from_space(sample_space=sample_space, resolution=[3, 3])
        cls.assertTrue(np.array_equiv(result, groundTruth))

    def test_grid_size_from_space(cls):
        """Should generate proper grid size from ranges."""

        xi = [np.linspace(-1, 1, 3), np.linspace(-1, 1, 3)]
        cls.assertEqual(grid_size_from_ranges(xi), 9)

    def test_grid_size_from_space(cls):
        """Should generate proper grid size from space."""

        sample_space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        cls.assertEqual(
            grid_size_from_space(sample_space=sample_space, resolution=3), 9
        )
        cls.assertEqual(
            grid_size_from_space(sample_space=sample_space, resolution=[3, 3]), 9
        )



r"""Backward in time stochastic optimal control.

The backward in time (dynamic programming) stochastic optimal control algorithm computes
the control actions working backward in time from the terminal time step to the current
time step. It computes a sequence of "value" functions, and then as the system
evolves forward in time, it chooses a control action that optimizes the value function,
rather than the actual cost.

The policy is specified as a sequence of stochastic kernels :math:`\pi = \lbrace
\pi_{0}, \pi_{1}, \ldots, \pi_{N-1} \rbrace`. Typically, the cost is formulated as an
additive cost structure, where at each time step the system incurs a *stage cost*
:math:`g_{t}(x_{t}, u_{t})`, and at the final time step, it incurs a *terminal cost*
:math:`g_{N}(x_{N})`.

.. math::

    \min_{\pi} \quad \mathbb{E} \biggl[ g_{N}(x_{N}) +
    \sum_{t=0}^{N-1} g_{t}(x_{t}, u_{t}) \biggr]

In dynamic programming, we solve the problem iteratively, by considering each time step
independently. We can do this by defining a sequence of *value functions* :math:`V_{0},
\ldots, V_{N}` that describe a type of "overall" cost at each time step, starting with
the terminal cost :math:`V_{N}(x) = g_{N}(x)`, and then substituting the subsequent
value function into the current one. Then, the policy is chosen to minimize (or
maximize, as is the convention for RL) the value functions.

.. math::

    V_{t}(x)
    = \max_{\pi_{t}} \quad \int_{\mathcal{U}} \int_{\mathcal{X}} g_{t}(x_{t}, u_{t})
    + V_{t+1}(y) Q(\mathrm{d} y \mid x, u) \pi_{t}(\mathrm{d} u \mid x)

Note:
    See :py:mod:`examples.benchmark_tracking_problem` for a complete example.

"""

from functools import partial

import numpy as np

from gym_socks.policies import BasePolicy
from gym_socks.algorithms.control.common import compute_solution

from gym_socks.kernel.metrics import rbf_kernel
from gym_socks.kernel.metrics import regularized_inverse

from gym_socks.sampling.transform import transpose_sample

from gym_socks.utils.batch import generate_batches

import logging
from gym_socks.utils.logging import ms_tqdm, _progress_fmt

logger = logging.getLogger(__name__)


def _compute_backward_recursion(
    Y,
    W,
    CXY,
    CUA,
    time_horizon=None,
    cost_fn=None,
    constraint_fn=None,
    value_functions=None,
    out=None,
    verbose=False,
):
    # Compute beta.
    logger.debug("Computing beta.")
    beta = np.einsum("ij,ik->ijk", CXY, CUA)

    pbar = ms_tqdm(
        total=time_horizon,
        bar_format=_progress_fmt,
        disable=False if verbose is True else True,
    )

    # set up empty array to hold value functions
    if out is None:
        out = np.zeros((time_horizon, len(Y)), dtype=np.float32)
    out[time_horizon - 1, :] = np.array(
        cost_fn(time=time_horizon - 1), dtype=np.float32
    )
    pbar.update()

    # run backwards in time and compute the safety probabilities
    for t in range(time_horizon - 2, -1, -1):

        Z = np.matmul(value_functions[t + 1, :], W)
        C = np.einsum("i,ijk->jk", Z, beta)

        V = np.zeros((len(Y),))

        if constraint_fn is not None:
            R = np.matmul(constraint_fn(time=t), W)
            D = np.einsum("i,ijk->jk", R, beta)

            for i in range(len(Y)):
                CA = C[i][D[i] <= 0]
                if CA.size == 0:
                    V[i] = np.min(C[i])
                else:
                    V[i] = np.min(CA)

        else:
            V = np.min(C, axis=1)

        # for i in range(len(Y)):
        #     sol = compute_solution(C[i], D[i])
        #     # idx = np.argmax(sol)
        #     # V =
        #     V[i] = sol

        out[t, :] = cost_fn(time=t) + V

        pbar.update()

    pbar.close()

    return out


def _compute_backward_recursion_batch(
    Y,
    W,
    CXY,
    CUA,
    time_horizon=None,
    cost_fn=None,
    constraint_fn=None,
    value_functions=None,
    out=None,
    batch_size=None,
    verbose=False,
):
    pbar = ms_tqdm(
        total=time_horizon,
        bar_format=_progress_fmt,
        disable=False if verbose is True else True,
    )

    # set up empty array to hold value functions
    if out is None:
        out = np.zeros((time_horizon + 1, len(Y)), dtype=np.float32)
    out[time_horizon - 1, :] = np.array(
        cost_fn(time=time_horizon - 1), dtype=np.float32
    )
    pbar.update()

    # run backwards in time and compute the safety probabilities
    for t in range(time_horizon - 2, -1, -1):

        Z = np.matmul(value_functions[t + 1, :], W)
        # C = np.empty((1, len(Y)))
        C = np.empty_like(CUA)

        # if constraint_fn is not None:
        # D = np.empty((1, len(Y)))
        D = np.empty_like(CUA)

        for batch in generate_batches(num_elements=len(Y), batch_size=batch_size):

            beta = np.einsum("ij,ik->ijk", CXY[:, batch], CUA)
            C[batch, :] = np.einsum("i,ijk->jk", Z, beta)

            if constraint_fn is not None:
                R = np.matmul(constraint_fn(time=t), W)
                D[batch, :] = np.einsum("i,ijk->jk", R, beta)

        # V = np.zeros((len(Y),))
        # for i in range(len(Y)):
        #     # sol = compute_solution(C[i], D[i])
        #     # idx = np.argmax(sol)
        #     # # print(f"V shape {V.shape}")
        #     # # print(f"V shape {V[i].shape}")
        #     # # print(f"C chape {C.shape}")
        #     # V[i] = C[:, idx]

        #     CA = C[i][D[i] <= 0]
        #     if CA.size == 0:
        #         V[i] = np.min(C[i])
        #     else:
        #         V[i] = np.min(CA)

        #     #     out[t, :] = cost_fn(time=t) + V

        #     # else:

        #     #     Z = np.matmul(value_functions[t + 1, :], W)

        #     #     C = np.zeros_like(CUA)

        #     #     for batch in generate_batches(num_elements=len(Y), batch_size=batch_size):

        #     #         beta = np.einsum("ij,ik->ijk", CXY[:, batch], CUA)
        #     #         C[batch, :] = np.einsum("i,ijk->jk", Z, beta)

        #     #     V = np.min(C, axis=1)

        if constraint_fn is not None:

            V = np.zeros((len(Y),))

            for i in range(len(Y)):

                CA = C[i][D[i] <= 0]
                if CA.size == 0:
                    V[i] = np.min(C[i])
                else:
                    V[i] = np.min(CA)
        else:
            V = np.min(C, axis=1)

        out[t, :] = cost_fn(time=t) + V

        pbar.update()

    pbar.close()

    return out


def kernel_control_bwd(
    S: np.ndarray,
    A: np.ndarray,
    time_horizon: int = None,
    cost_fn=None,
    constraint_fn=None,
    heuristic: bool = False,
    regularization_param: float = None,
    kernel_fn=None,
    batch_size: int = None,
    verbose: bool = True,
):
    """Stochastic optimal control policy backward in time.

    Computes the optimal policy using dynamic programming. The solution computes an approximation of the value functions starting at the terminal time and working backwards. Then, when the policy is evaluated, it moves forward in time, optimizing over the value functions and choosing the action which has the highest "value".

    Args:
        S: Sample taken iid from the system evolution.
        A: Collection of admissible control actions.
        cost_fn: The cost function. Should return a real value.
        constraint_fn: The constraint function. Should return a real value.
        heuristic: Whether to use the heuristic solution instead of solving the LP.
        regularization_param: Regularization prameter for the regularized least-squares
            problem used to construct the approximation.
        kernel_fn: The kernel function used by the algorithm.
        verbose: Whether the algorithm should print verbose output.

    """

    alg = KernelControlBwd(
        time_horizon=time_horizon,
        cost_fn=cost_fn,
        constraint_fn=constraint_fn,
        heuristic=heuristic,
        regularization_param=regularization_param,
        kernel_fn=kernel_fn,
        batch_size=batch_size,
        verbose=verbose,
    )
    alg.train(S, A)

    return alg


class KernelControlBwd(BasePolicy):
    """Stochastic optimal control policy backward in time.

    Computes the optimal policy using dynamic programming. The solution computes an approximation of the value functions starting at the terminal time and working backwards. Then, when the policy is evaluated, it moves forward in time, optimizing over the value functions and choosing the action which has the highest "value".

    Args:
        S: Sample taken iid from the system evolution.
        A: Collection of admissible control actions.
        cost_fn: The cost function. Should return a real value.
        constraint_fn: The constraint function. Should return a real value.
        heuristic: Whether to use the heuristic solution instead of solving the LP.
        regularization_param: Regularization prameter for the regularized least-squares
            problem used to construct the approximation.
        kernel_fn: The kernel function used by the algorithm.
        verbose: Whether the algorithm should print verbose output.

    """

    def __init__(
        self,
        time_horizon: int = None,
        cost_fn=None,
        constraint_fn=None,
        heuristic: bool = False,
        regularization_param: float = None,
        kernel_fn=None,
        batch_size: int = None,
        verbose: bool = True,
        *args,
        **kwargs,
    ):

        self.time_horizon = time_horizon

        self.cost_fn = cost_fn
        self.constraint_fn = constraint_fn

        self.heuristic = heuristic

        self.kernel_fn = kernel_fn
        self.regularization_param = regularization_param

        self.batch_size = batch_size

        self.verbose = verbose

    def _validate_params(self, S=None, A=None):

        if self.kernel_fn is None:
            self.kernel_fn = partial(rbf_kernel, sigma=0.1)

        if self.regularization_param is None:
            self.regularization_param = 1

        if S is None:
            raise ValueError("Must supply a sample.")

        if A is None:
            raise ValueError("Must supply a sample.")

        if self.cost_fn is None:
            raise ValueError("Must supply a cost function.")

        if self.constraint_fn is None:
            # raise ValueError("Must supply a constraint function.")
            self._constrained = False
        else:
            self._constrained = True

        if self.batch_size is not None:
            self._compute_backward_recursion_caller = partial(
                _compute_backward_recursion_batch, batch_size=self.batch_size
            )

        else:
            self._compute_backward_recursion_caller = _compute_backward_recursion

    def _validate_data(self, S):

        if S is None:
            raise ValueError("Must supply a sample.")

    def train(self, S: np.ndarray, A: np.ndarray):
        """Train the algorithm.

        Args:
            S: Sample taken iid from the system evolution.
            A: Collection of admissible control actions.

        Returns:
            An instance of the KernelControlFwd algorithm class.

        """

        self._validate_data(S)
        self._validate_data(A)
        self._validate_params(S, A)

        X, U, Y = transpose_sample(S)
        X = np.array(X)
        U = np.array(U)
        Y = np.array(Y)

        self.X = X

        A = np.array(A)

        self.A = A

        logger.debug("Computing matrix inverse.")
        self.W = regularized_inverse(
            X,
            U=U,
            kernel_fn=self.kernel_fn,
            regularization_param=self.regularization_param,
        )

        logger.debug("Computing covariance matrices.")
        self.CUA = self.kernel_fn(U, A)
        CXY = self.kernel_fn(X, Y)

        logger.debug("Computing cost function.")
        self.cost_fn = partial(self.cost_fn, state=Y)

        logger.debug("Computing constraint function.")
        if self.constraint_fn is not None:
            self.constraint_fn = partial(self.constraint_fn, state=Y)

        logger.debug("Computing value functions.")
        value_functions = np.empty((self.time_horizon, len(Y)))
        self.value_functions = self._compute_backward_recursion_caller(
            Y,
            self.W,
            CXY,
            self.CUA,
            time_horizon=self.time_horizon,
            cost_fn=self.cost_fn,
            constraint_fn=self.constraint_fn,
            value_functions=value_functions,
            out=value_functions,
            verbose=True,
        )

    def __call__(self, time=0, state=None, *args, **kwargs):

        if state is None:
            print("Must supply a state to the policy.")
            return None

        T = np.array(state)

        # Compute covariance matrix.
        CXT = self.kernel_fn(self.X, T)

        # Compute beta.
        betaXT = self.W @ (CXT * self.CUA)

        # Compute cost vector.
        C = np.matmul(np.array(self.value_functions[time], dtype=np.float32), betaXT)

        # Compute constraint vector.
        D = None
        if self.constraint_fn is not None:
            D = np.matmul(
                np.array(self.constraint_fn(time=time), dtype=np.float32), betaXT
            )

        # Compute the solution to the LP.
        logger.debug("Computing solution to the LP.")
        sol = compute_solution(C, D, heuristic=self.heuristic)
        idx = np.argmax(sol)
        return np.array(self.A[idx], dtype=np.float32)



import unittest
from unittest.mock import patch

import gym

import gym_socks.envs

from gym_socks.policies import BasePolicy

from gym_socks.policies import ConstantPolicy
from gym_socks.policies import RandomizedPolicy
from gym_socks.policies import ZeroPolicy

import numpy as np


class TestBasePolicy(unittest.TestCase):
    @classmethod
    @patch("gym_socks.policies.policy.BasePolicy.__abstractmethods__", set())
    def setUpClass(cls):
        cls.policy = BasePolicy()

    def test_default_policy_not_implemented(cls):

        with cls.assertRaises(NotImplementedError):
            cls.policy()


class TestConstantPolicy(unittest.TestCase):
    @classmethod
    @patch("gym_socks.envs.dynamical_system.DynamicalSystem.__abstractmethods__", set())
    def setUpClass(cls):
        cls.action_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
        )

    def test_constant_policy_returns_constants(cls):
        policy = ConstantPolicy(action_space=cls.action_space, constant=1)
        cls.assertEqual(policy(), [1])

        policy = ConstantPolicy(action_space=cls.action_space, constant=5.0)
        cls.assertEqual(policy(), [5.0])


class TestZeroPolicy(unittest.TestCase):
    @classmethod
    @patch("gym_socks.envs.dynamical_system.DynamicalSystem.__abstractmethods__", set())
    def setUpClass(cls):
        cls.action_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
        )

    def test_zero_policy_returns_zero(cls):
        policy = ZeroPolicy(action_space=cls.action_space)
        cls.assertEqual(policy(), [0])



import unittest

import gym

from gym_socks.utils.batch import generate_batches

import numpy as np


class TestGenerateBatches(unittest.TestCase):
    """Test generate_batches."""

    def test_generate_batches(cls):
        """Test generate batches."""
        idx = np.arange(10)

        batches = generate_batches(num_elements=10, batch_size=2)
        for i, batch in enumerate(batches):
            cls.assertTrue(len(idx[batch]) == 2)
            cls.assertTrue(np.all(np.equal(idx[batch], idx[i * 2 : i * 2 + 2])))

        batches = generate_batches(num_elements=10, batch_size=5)
        for i, batch in enumerate(batches):
            cls.assertTrue(len(idx[batch]) == 5)
            cls.assertTrue(np.all(np.equal(idx[batch], idx[i * 5 : i * 5 + 5])))

    def test_batch_larger_than_set(cls):
        """Ensure batches past the maximum size work properly."""

        idx = np.arange(10)

        batches = generate_batches(num_elements=14, batch_size=7)
        batches.send(None)
        batch = batches.send(None)
        cls.assertTrue(len(idx[batch]) == 3)
        cls.assertTrue(np.all(np.equal(idx[batch], idx[-3:])))

        batches = generate_batches(num_elements=10, batch_size=14)
        batch = batches.send(None)
        cls.assertTrue(len(idx[batch]) == 10)
        cls.assertTrue(np.all(np.equal(idx[batch], idx[:])))



from gym_socks.policies.policy import BasePolicy

from gym_socks.policies.policy import ConstantPolicy
from gym_socks.policies.policy import RandomizedPolicy
from gym_socks.policies.policy import ZeroPolicy


__all__ = [
    "BasePolicy",
    "ConstantPolicy",
    "RandomizedPolicy",
    "ZeroPolicy",
]



def generate_batches(num_elements, batch_size):
    """Generate batches.

    Batch generation function to split a list into smaller batches (slices). Generates
    `slice` objects, which can be iterated over in a for loop.

    Args:
        num_elements: Length of the list.
        batch_size: Maximum size of the batches. If the last batch is smaller than
            `batch_size`, then the final batch will be the length of the remaining
            elements.

    Yields:
        A slice object.

    """

    start = 0

    for _ in range(num_elements // batch_size):
        end = start + batch_size

        yield slice(start, end)

        start = end

    if start < num_elements:
        yield slice(start, num_elements)



import unittest
from unittest.mock import patch

import gym

from gym_socks.envs.tora import TORAEnv

import numpy as np


class TestToraSystem(unittest.TestCase):
    def test_set_damping_coefficient(cls):
        env = TORAEnv()

        cls.assertEqual(TORAEnv._damping_coefficient, 0.1)
        cls.assertEqual(env.damping_coefficient, 0.1)

        env.damping_coefficient = 0.5
        cls.assertEqual(env.damping_coefficient, 0.5)



"""Control policies.

Note:
    Policies ccan be either time-invariant or time-varying, and can be either open- or
    closed-loop. Thus, the arguments to the :py:meth:`__call__` method should allow for
    ``time`` and ``state`` to be specified (if needed), and should be optional kwargs::

        >>> def __call__(self, time=None, state=None):
        ...     ...

"""

from abc import ABC, abstractmethod

from gym import Space

import numpy as np


class BasePolicy(ABC):
    """Base policy class.

    This class is **abstract**, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from :py:class:`BasePolicy`.

    The :py:meth:`__call__` method is the main point of entry for the policy classes.
    All subclasses must implement a py:meth:`__call__` method. This makes the class
    callable, so that policies can be evaluated as::

        action = policy(state)

    """

    action_space = None
    state_space = None

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """Evaluate the policy.

        Returns:
            An action in the action space.

        """
        raise NotImplementedError


class RandomizedPolicy(BasePolicy):
    """Randomized policy.

    A policy which returns a random control action.

    Args:
        action_space: The action space of the system.

    """

    def __init__(self, action_space: Space = None):
        if action_space is not None:
            self.action_space = action_space
        else:
            raise ValueError("action space must be provided")

    def __call__(self, *args, **kwargs):
        return self.action_space.sample()


class ConstantPolicy(BasePolicy):
    """Constant policy.

    A policy which returns a constant control action.

    Args:
        action_space: The action space of the system.
        constant: The constant value returned by the policy.

    """

    def __init__(self, action_space: Space = None, constant=0):
        if action_space is not None:
            self.action_space = action_space
        else:
            raise ValueError("action space must be provided")

        self._constant = constant

    def __call__(self, *args, **kwargs):
        return np.full(
            self.action_space.shape,
            self._constant,
            dtype=self.action_space.dtype,
        )


class ZeroPolicy(ConstantPolicy):
    """Zero policy.

    A policy which returns a constant (zero) control action.

    Args:
        action_space: The action space of the system.

    """

    def __init__(self, action_space: Space = None):
        if action_space is not None:
            self.action_space = action_space
        else:
            raise ValueError("action space must be provided")

        self._constant = 0



from abc import ABC, abstractmethod
from collections.abc import MutableSequence

from gym_socks.envs.core import BaseDynamicalObject

import numpy as np
from scipy.integrate import solve_ivp


class _WorldObjectMeta(type):
    """_WorldObject meta class.

    The meta class defines a virtual interface for world objects, which must implement
    the following methods: ``step``, ``reset``, ``render``, ``close``, and ``seed``.

    """

    def __instancecheck__(cls, instance):
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        return (
            hasattr(subclass, "step")
            and callable(subclass.step)
            and hasattr(subclass, "reset")
            and callable(subclass.reset)
            and hasattr(subclass, "render")
            and callable(subclass.render)
            and hasattr(subclass, "close")
            and callable(subclass.close)
            and hasattr(subclass, "seed")
            and callable(subclass.seed)
        )


class _WorldObject(metaclass=_WorldObjectMeta):
    """Virtual interface for world objects.

    Example:

        >>> class DummyObject(object):
        ...     def step(self):
        ...         pass
        ...
        ...     def reset(self):
        ...         pass
        ...
        ...     def render(self):
        ...         pass
        ...
        ...     def close(self):
        ...         pass
        ...
        ...     def seed(self):
        ...         pass
        ...
        >>> dummy = DummyObject()
        >>> isinstance(DummyObject, _WorldObject)

    """

    pass


class World(MutableSequence):
    """World.

    The ``World`` class is essentially a ``MutableSequence`` (such as a list), of
    objects contained within the world. The objects contained within the world must be
    of type ``_WorldObject``, meaning the implement the ``step``, ``reset``, ``render``,
    ``close``, and ``seed`` methods. If an object that does not implement these methods
    is added, an assertion error will be thrown.

    It can be used in much the same was as a list, for example::

        >>> world = World()
        >>> world.append(item)
        >>> world[1] = item
        >>> world += [item]

    The ``World`` is primarily used for keeping track of multiple objects, such as
    obstacles, which are contained within the world environment and simulated together.
    While it can track ``DynamicalSystem``, it is moreso intended to track uncontrolled
    or fully autonomous systems or objects, and to synchronize the simulation time.

    In addition, it implements the following methods, which are applied to all objects
    in the world environment: ``step``, ``reset``, ``render``, ``close``, and ``seed``.

    E.g., the ``step`` method calls the ``step`` method for each item in the world. The
    order in which the items are iterated over is the same as the order of the list. If
    the functions return a value, the results are given in a list.

    """

    _time_horizon = 1

    def __init__(self) -> None:
        self._objects = []

    @property
    def time_horizon(self):
        """Time horizon for the simulation."""
        return self._time_horizon

    @time_horizon.setter
    def time_horizon(self, value: int):
        msg = "Horizon must be an integer."
        assert value >= 0 and isinstance(value, (int, np.integer)), msg

        self._time_horizon = int(value)

    @staticmethod
    def _check_item(item: object):
        """Checks whether an item is a ``_WorldObject``.

        Raises:
            ValueError: If item is not an instance of ``_WorldObject``.

        """

        if not isinstance(item, _WorldObject):
            raise ValueError("invalid world object")

    def __getitem__(self, _index: int):
        return self._objects[_index]

    def __setitem__(self, _index: int, _object: _WorldObject):
        self._check_item(_object)
        self._objects[_index] = _object

    def __delitem__(self, _index: int):
        del self._objects[_index]

    def __len__(self):
        return len(self._objects)

    def insert(self, _index: int, _object: _WorldObject):
        self._check_item(_object)
        self._objects.insert(_index, _object)

    def step(self, time: int = None) -> list:
        """Advances the simulation forward one time step.

        Args:
            time: The simulation time.

        Returns:
            The result of each step function in a list.

        """

        result = []
        for item in self._objects:
            result.append(item.step(time=time))

        return result

    def reset(self) -> list:
        """Reset the world to a random initial condition.

        Returns:
            The result of each reset function in a list.

        """

        result = []
        for item in self._objects:
            result.append(item.reset())

        return result

    def render(self, mode: str = "human"):
        for item in self._objects:
            item.render(mode)

    def close(self):
        for item in self._objects:
            item.close()

    def seed(self, seed: int = None) -> list:
        """Sets the seed of the random number generator.

        This is primarily useful for objects which incorporate some sort of
        stochasticity to ensure repeatability.

        Args:
            seed: Integer value representing the random seed.

        Returns:
            The seed of the RNG for each object in a list.

        """

        result = []
        for item in self._objects:
            result.append(item.seed(seed=seed))

        return result



"""Sampling methods.

This module contains a collection of sampling methods. The core principle is to define a
function that returns a single observation (either via return or yield) from a
probability measure. Then, using the decorator ``sample_generator``, a function that
returns a single observation can be converted into a generator, that can then be sampled
using ``islice``.

Example:

    Sample the stochastic kernel of a dynamical system (i.e. the state transition
    probability kernel).

        >>> from gym_socks.envs.integrator import NDIntegratorEnv
        >>> from gym_socks.policies import RandomizedPolicy
        >>> env = NDIntegratorEnv(2)
        >>> sample_space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        >>> sampler = default_sampler(
        ...     system=env, policy=RandomizedPolicy(env), sample_space=sample_space
        ... )
        >>> S = sample(sampler=sampler, sample_size=100)

The main reason for this setup is to allow for "observation functions" which have
different structures, e.g.

* functions that ``return`` an observation,
* infinite generators that ``yield`` observations, and
* finite generators that ``yield`` observations.

"""

from gym_socks.sampling.sample import sample_generator

from gym_socks.sampling.sample import sample

from gym_socks.sampling.sample import default_sampler
from gym_socks.sampling.sample import default_trajectory_sampler
from gym_socks.sampling.sample import grid_sampler
from gym_socks.sampling.sample import random_sampler

from gym_socks.sampling.sample import repeat

__all__ = [
    "sample_generator",
    "sample",
    "default_sampler",
    "default_trajectory_sampler",
    "grid_sampler",
    "random_sampler",
    "repeat",
]



import gym

import numpy as np

from gym_socks.utils import indicator_fn


def _fht_step(Y, V, constraint_set, target_set):
    r"""First-hitting time problem backward recursion step.

    This function implements the backward recursion step for the first-hitting time
    problem, given by:

    .. math::

        V_{t}(x) = 1_{\mathcal{T}}(x)
        + 1_{\mathcal{K} \backslash \mathcal{T}}(x) \mathbb{E}[V_{t+1}]

    Args:
        Y: Vector of data points in the sample.
        V: Vector of coefficients from the kernel-based approximation.
        constraint_set: A space or function which determines whether a point is in the
            constraint set. This can be either a space from gym.spaces or a function
            which acts as a 0/1 classifier. See the `indicator_fn` implementation for
            more information.
        target_set: A space or function which determines whether a point is in the
            target set. This can be either a space from gym.spaces or a function which
            acts as a 0/1 classifier. See the `indicator_fn` implementation for more
            information.

    Returns:
        The kernel-based approximation of the value function.

    """

    Y_in_constraint_set = indicator_fn(Y, constraint_set)
    Y_in_target_set = indicator_fn(Y, target_set)

    return Y_in_target_set + (Y_in_constraint_set & ~Y_in_target_set) * V


def _tht_step(Y, V, constraint_set, target_set):
    r"""Terminal-hitting time problem backward recursion step.

    This function implements the backward recursion step for the terminal-hitting time
    problem, given by:

    .. math::

        V_{t}(x) = 1_{\mathcal{K}}(x) \mathbb{E}[V_{t+1}]

    Args:
        Y: Vector of data points in the sample.
        V: Vector of coefficients from the kernel-based approximation.
        constraint_set: A space or function which determines whether a point is in the
            constraint set. This can be either a space from gym.spaces or a function
            which acts as a 0/1 classifier. See the `indicator_fn` implementation for
            more information.
        target_set: This argument is unused in the terminal-hitting time problem. The
            reson it is allowed is to avoid extra if statements in the implementation.

    Returns:
        The kernel-based approximation of the value function.

    """

    Y_in_constraint_set = indicator_fn(Y, constraint_set)

    return Y_in_constraint_set * V


def generate_tube(time_horizon: int, shape, low, high):
    """Generate a stochastic reachability tube using config.

    This function computes a stochastic reachability tube using the tube configuration.

    Args:
        env: The dynamical system model.
        bounds: The bounds of the tube. Specified as a dictionary.

    Returns:
        A list of spaces indexed by time.

    """

    tube_lb = bounds["lower_bound"]
    tube_ub = bounds["upper_bound"]

    tube = []
    for i in range(time_horizon):
        tube_t = gym.spaces.Box(
            low[i],
            high[i],
            shape=shape,
            dtype=np.float32,
        )
        tube.append(tube_t)

    return tube



from abc import ABC, abstractmethod

from gym_socks.envs.core import BaseDynamicalObject


class BaseObstacle(BaseDynamicalObject, ABC):
    """Base obstacle class.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from BaseObstacle.

    All obstacles inherit from BaseDynamicalObject, meaning they must implement the
    `step`, `reset`, `render`, `close`, and `seed` methods in their custom
    implementations. This is necessary so that the `World` class in
    `gym_socks.envs.world` can register them as objects in the world environment.

    In addition, all obstacles must implement an `intersects` method, which returns true
    if the state of the dynamical system intersects with the obstacle. For instance,
    this could identify a collision, an unsafe region, or a non-violable constraint.

    """

    @abstractmethod
    def intersects(self, time, state):
        """Check if the state intersects the obstacle.

        This function is used to check whether or not the state of a dynamical system intersects the given obstacle. It should return True if the system intersects the obstacle and False otherwise.

        Args:
            time: The time step of the simulation.
            state: The state of the dynamical system.

        Returns:
            A boolean indicator.

        """

        raise NotImplementedError



import unittest
from unittest.mock import patch

import gym

from gym_socks.envs.dynamical_system import DynamicalSystem

from gym_socks.envs import NDIntegratorEnv
from gym_socks.envs import NDPointMassEnv
from gym_socks.envs import NonholonomicVehicleEnv
from gym_socks.envs import PlanarQuadrotorEnv
from gym_socks.envs import CWH4DEnv
from gym_socks.envs import CWH6DEnv
from gym_socks.envs import QuadrotorEnv
from gym_socks.envs import TORAEnv

import numpy as np

system_list = [
    NDIntegratorEnv(1),
    NDIntegratorEnv(2),
    NDIntegratorEnv(3),
    NDIntegratorEnv(4),
    NDPointMassEnv(1),
    NDPointMassEnv(2),
    NDPointMassEnv(3),
    NDPointMassEnv(4),
    NonholonomicVehicleEnv(),
    PlanarQuadrotorEnv(),
    CWH4DEnv(),
    CWH6DEnv(),
    QuadrotorEnv(),
    TORAEnv(),
]


class TestEnvironmentsRun(unittest.TestCase):
    def test_envs_run(cls):
        """Assert that environments can run."""

        for env in system_list:
            with cls.subTest(msg=f"Testing with {type(env)}."):

                try:
                    obs = env.reset()

                    for t in range(5):

                        # get action
                        action = env.action_space.sample()

                        # apply action
                        obs, cost, done, _ = env.step(time=t, action=action)

                except Exception as e:
                    cls.fail(f"Simulating system {type(env)} raised an exception.")


class TestDynamicalSystem(unittest.TestCase):
    """Dynamical system tests."""

    @patch("gym_socks.envs.dynamical_system.DynamicalSystem.__abstractmethods__", set())
    def setUp(cls):
        cls.env = DynamicalSystem()

        cls.env.observation_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
        )
        cls.env.state_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
        )
        cls.env.action_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
        )

    # def test_system_num_time_steps(cls):
    #     """System returns correct number of time steps."""

    #     cls.env.time_horizon = 1
    #     cls.env.sampling_time = 0.1

    #     # Note that this is not an error, and has to do with floating point precision.
    #     # See: https://docs.python.org/3/tutorial/floatingpoint.html
    #     cls.assertEqual(cls.env.num_time_steps, 9)
    #     cls.assertEqual(cls.env.time_horizon, 1)
    #     cls.assertEqual(cls.env.sampling_time, 0.1)

    #     cls.env.time_horizon = 10.0
    #     cls.env.sampling_time = 1.0

    #     # Note that this is not an error, and has to do with floating point precision.
    #     # See: https://docs.python.org/3/tutorial/floatingpoint.html
    #     cls.assertEqual(cls.env.num_time_steps, 10)
    #     cls.assertEqual(cls.env.time_horizon, 10)
    #     cls.assertEqual(cls.env.sampling_time, 1.0)

    #     cls.env.time_horizon = 5
    #     cls.env.sampling_time = 1

    #     cls.assertEqual(cls.env.num_time_steps, 5)
    #     cls.assertEqual(cls.env.time_horizon, 5)
    #     cls.assertEqual(cls.env.sampling_time, 1)

    # def test_dims(cls):
    #     """State and action dims should match spaces."""
    #     cls.assertEqual(cls.env.state_dim, (1,))
    #     cls.assertEqual(cls.env.action_dim, (1,))

    #     cls.env.observation_space = gym.spaces.Box(
    #         low=-1, high=1, shape=(1,), dtype=np.float32
    #     )

    #     cls.assertEqual(cls.env.state_dim, (1,))

    def test_reset_returns_valid_state(cls):
        """Reset should return a valid state."""
        cls.env.reset()
        cls.assertTrue(cls.env.observation_space.contains(cls.env.state))

    def test_default_dynamics_not_implemented(cls):

        with cls.assertRaises(NotImplementedError):
            # state = cls.env.reset()
            # action = cls.env.action_space.sample()
            # disturbance = cls.env.generate_disturbance(0, state, action)
            cls.env.dynamics(0, None, None, None)

    # def test_default_close_not_implemented(cls):

    #     with cls.assertRaises(NotImplementedError):
    #         cls.env.close()

    def test_default_render_not_implemented(cls):

        with cls.assertRaises(NotImplementedError):
            cls.env.render()



"""Sampling methods."""

from inspect import isgeneratorfunction
from functools import partial, wraps
from itertools import islice

import gym

from gym_socks.envs.dynamical_system import DynamicalSystem


def sample_generator(fun):
    """Sample generator decorator.

    Converts a sample function into a generator function. Any function that returns a
    single observation (as a tuple) can be converted into a sample generator.

    Args:
        fun: Sample function that returns or yields an observation.

    Returns:
        A function that can be used to ``islice`` a sample from the sample generator.

    Example:

        >>> from itertools import islice
        >>> from gym_socks.envs.sample import sample_generator
        >>> @sample_generator
        ... def custom_sampler(env, policy, sample_space):
        ...     env.state = sample_space.sample()
        ...     action = policy(state=state)
        ...     next_state, *_ = env.step(action)
        ...     yield (env.state, action, next_state)
        >>> S = list(islice(custom_sampler(), 100))

    """

    @wraps(fun)
    def _wrapper(*args, **kwargs):
        if isgeneratorfunction(fun):
            while True:
                yield from fun(*args, **kwargs)
        else:
            while True:
                yield fun(*args, **kwargs)

    return _wrapper


@sample_generator
def random_sampler(sample_space: gym.Space):
    """Random sampler.

    Returns a random sample taken from the space. See :py:obj:`gym.spaces.Box` for more
    information on the distributions used for sampling.

    Args:
        sample_space: The space to sample from.

    Yields:
        A sample from the ``sample_space``.

    """

    yield sample_space.sample()


@sample_generator
def grid_sampler(grid: list):
    """Grid sampler.

    Returns a sample arranged on a uniformly-spaced grid. Use
    :py:func:`make_grid_from_ranges` or :py:func:`make_grid_from_space` from
    :py:mod:`gym_socks.utils.grid` to generate a grid of points.

    Args:
        grid: The grid of points.

    Yields:
        A sample from the ``sample_space``.

    """

    for item in grid:
        yield item


@sample_generator
def repeat(sampler, num: int):
    """Repeat sampler.

    Repeats the output of a sample generator ``num`` times.

    Args:
        sampler: The sample generator function.
        num: The number of times to repeat a sample.

    Yields:
        A repeated sample.

    """

    for item in sampler:
        for i in range(num):
            yield item


def default_sampler(
    state_sampler=None,
    action_sampler=None,
    env: DynamicalSystem = None,
):
    """Default trajectory sampler.

    Args:
        state_sampler: The state space sampler.
        action_sampler: The action space sampler.
        env: The system to sample from.

    Returns:
        A generator function that yields system observations as tuples.

    """

    @sample_generator
    def _sample_generator():
        state = next(state_sampler)
        action = next(action_sampler)

        env.state = state
        next_state, *_ = env.step(action=action)

        yield (state, action, next_state)

    return _sample_generator


def default_trajectory_sampler(
    state_sampler=None,
    action_sampler=None,
    env: DynamicalSystem = None,
    time_horizon: int = 1,
):
    """Default trajectory sampler.

    Args:
        state_sampler: The state space sampler.
        action_sampler: The action space sampler.
        env: The system to sample from.
        time_horizon: The time horizon to simulate over.

    Returns:
        A generator function that yields system observations as tuples.

    """

    @sample_generator
    def _sample_generator():
        state = next(state_sampler)

        state_sequence = []
        action_sequence = []

        env.state = state

        time = 0
        for t in range(time_horizon):
            action = next(action_sampler)
            next_state, *_ = env.step(time=t, action=action)

            state_sequence.append(next_state)
            action_sequence.append(action)

            time += 1

        yield (state, action_sequence, state_sequence)

    return _sample_generator


def sample(sampler=None, sample_size: int = None, *args, **kwargs):
    """Generate a sample using the sample generator.

    Args:
        sampler: Sample generator function.
        sample_size: Size of the sample.

    Returns:
        list of tuples

    """

    if sampler is None:
        raise ValueError("Must supply a sample function.")

    if sample_size is None:
        raise ValueError("Must supply a sample size.")

    if not isinstance(sample_size, int) or sample_size < 1:
        raise ValueError("Sample size must be a positive integer.")

    return list(islice(sampler(*args, **kwargs), sample_size))



from abc import ABC, abstractmethod

import numpy as np
from numpy.linalg import inv
from scipy.linalg import block_diag

from functools import partial

from gym_socks.algorithms.base import RegressorMixin

from sklearn.preprocessing import normalize


def _regression_score(y_true, y_pred):
    return np.abs(np.asarray(y_true) - np.asarray(y_pred)).mean()


class ConditionalEmbedding(RegressorMixin):
    r"""Conditional distribution embedding.

    Computes the conditional distribution embedding using sample data.

    This class is primarily used internally to compute the regularized matrix inverse
    and predict the values of a function.

    .. math::

        m(x) = f^{\top} (G + \lambda M I)^{-1} K

    Args:
        regularization_param: The regularization parameter :math:`\lambda`.

    """

    def __init__(
        self,
        regularization_param: float = None,
    ):
        self.regularization_param = regularization_param

        self._W = None

    def fit(self, G: np.ndarray):
        r"""Fit the model.

        Computes the matrix inverse :math:`W = (G + \lambda M I)^{-1}`.

        Args:
            G: The (kernel) gram matrix. Usually computed as ``kernel_fn(X_train)``.

        Returns:
            self

        """

        s = np.shape(G)
        if len(s) != 2 and s[0] != s[1]:
            raise ValueError("Gram matrix must be square.")

        if self.regularization_param is None:
            self.regularization_param = 1 / (len(G) ** 2)
        else:
            assert (
                self.regularization_param > 0
            ), "regularization_param must be a strictly positive real value."

        I = np.zeros(np.shape(G))
        np.fill_diagonal(I, self.regularization_param * len(G))
        self._W = G + I
        return self

    def predict(self, y: np.ndarray, K: np.ndarray):
        r"""Predict the values.

        Computes the product :math:`y^{\top} W K`.

        Args:
            y: The output values of the function.
            K: The kernel matrix. Usually computed as ``kernel_fn(X_train, X_test)``.

        Returns:
            An ndarray of predicted y values.

        """

        return (y.T @ np.linalg.solve(self._W, K)).T

    def score(self, y_test: np.ndarray, K: np.ndarray, y_train: np.ndarray):
        return _regression_score(y_test, self.predict(y_train, K))




from abc import ABC, abstractmethod

import gym
from gym.utils import seeding

import numpy as np
from scipy.integrate import solve_ivp

from gym_socks.envs.core import BaseDynamicalObject


class DynamicalSystem(BaseDynamicalObject, ABC):
    r"""Base class for dynamical system models.

    Bases: :py:class:`gym_socks.envs.core.BaseDynamicalObject`, :py:obj:`abc.ABC`

    This class is **abstract**, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from :py:class:`DynamicalSystem`, and
    define a custom :py:meth:`dynamics` function.

    Example:
        ::

            import gym
            import numpy as np
            from gym_socks.envs.dynamical_system import DynamicalSystem
            class CustomDynamicalSystem(DynamicalSystem):

                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                    state_dim = 2  # 2-D state and observation space.
                    action_dim = 1  # 1-D action space.

                    self.observation_space = gym.spaces.Box(
                        low=-np.inf, high=np.inf, shape=(state_dim,), dtype=np.float32
                    )
                    self.state_space = gym.spaces.Box(
                        low=-np.inf, high=np.inf, shape=(state_dim,), dtype=np.float32
                    )
                    self.action_space = gym.spaces.Box(
                        low=-np.inf, high=np.inf, shape=(action_dim,), dtype=np.float32
                    )

                    self.state = None

                    self.seed(seed=seed)

                def dynamics(self, time, state, action, disturbance):
                    ...

    Important:
        The dynamics function (defined by you) returns :math:`\dot{x}`, and the
        system is integrated using :py:obj:`scipy.integrate.solve_ivp` in the
        :py:meth:`step` function to determine the state at the next time instant
        and discretize the system in time.

        To specify discrete time dynamics explicitly (for instance with linear
        dynamics such as :math:`x_{t+1} = A x_{t} + B x_{t} + w_{t}` where :math:`A`
        and :math:`B` are known, but the ODEs are difficult to write or are
        time-varying), override the :py:meth:`step` function.

        For example::

            import gym
            import numpy as np
            from gym_socks.envs.dynamical_system import DynamicalSystem
            class CustomDynamicalSystem(DynamicalSystem):

                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                    state_dim = 2  # 2-D state and observation space.
                    action_dim = 1  # 1-D action space.

                    self.observation_space = gym.spaces.Box(
                        low=-np.inf, high=np.inf, shape=(state_dim,), dtype=np.float32
                    )
                    self.state_space = gym.spaces.Box(
                        low=-np.inf, high=np.inf, shape=(state_dim,), dtype=np.float32
                    )
                    self.action_space = gym.spaces.Box(
                        low=-np.inf, high=np.inf, shape=(action_dim,), dtype=np.float32
                    )

                    self.state = None

                    self.seed(seed=seed)

                def step(self, time, action):

                    disturbance = self.generate_disturbance(time, self.state, action)
                    self.state = self.dynamics(time, self.state, action, disturbance)
                    obs = self.generate_observation(time, self.state, action)

                    return obs, 0, False, {}

                def compute_state_matrix(time):
                    ...

                def compute_input_matrix(time):
                    ...

                def dynamics(self, time, state, action, disturbance):

                    A = compute_state_matrix(time)
                    B = compute_input_matrix(time)

                    return A @ state + B @ action + disturbance

    Important:
        The state space and input space are assumed to be :math:`\mathbb{R}^{n}` and
        :math:`\mathbb{R}^{m}`, respectively, where :math:`n` and :math:`m` are
        specified by ``state_dim`` and ``action_dim`` above.

        In addition, the default :py:meth:`step` function solves an initial value
        problem via :py:obj:`scipy.integrate.solve_ivp` to compute the state of the
        system at the next time step. Thus, in order to use discrete spaces, you will
        need to override the :py:meth:`step` function.

        See :py:obj:`gym.spaces` for more information on the different available spaces.

    The system can then be simulated using the standard gym environment.

    Example:

        >>> env = CustomDynamicalSystem()
        >>> env.reset()
        >>> for t in range(10):
        ...     action = env.action_space.sample()
        ...     obs, *_ = env.step(action)
        ...
        >>> env.close()

    """

    observation_space = None
    """The space of system observations.

    Note:
        The observation space typically only differs from the state space if the system
        is partially observable. If this is the case, :py:meth:`generate_observation`
        should be defined to return an element of the observation space.

    """
    action_space = None
    """The action (input) space of the system."""
    state_space = None
    """The state space of the system.

    Note:
        By convention, in controls theory the state space of the system is the set of
        all possible states that the system can have. OpenAI Gym's convention is to
        ignore the underlying state space, opting to use only the
        :py:attr:`observation_space`.

    """

    _euler = False

    _sampling_time = 0.1

    @property
    def sampling_time(self):
        """Sampling time, in seconds."""
        return self._sampling_time

    @sampling_time.setter
    def sampling_time(self, value):
        self._sampling_time = value

    def generate_disturbance(self, time, state, action):
        """Generate a disturbance.

        Note:
            Override :py:meth:`generate_disturbance` in subclasses to modify the
            disturbance properties, such as the scale or distribution.

        """

        return self._np_random.standard_normal(size=self.state_space.shape)

    @abstractmethod
    def dynamics(self, time, state, action, disturbance):
        """Dynamics for the system.

        .. math::

            \dot{x} = f(x, u, w)

        Args:
            time: The time variable.
            state: The state of the system at the current time step.
            action: The control action applied at the current time step.
            disturbance: A realization of a random variable representing process noise.

        Returns:
            The state of the system at the next time step.

        """
        raise NotImplementedError

    def generate_observation(self, time, state, action):
        """Generate an observation from the system.

        .. math::

            y = h(x, u, v)

        Args:
            time: The time variable.
            state: The state of the system at the current time step.
            action: The control action applied at the current time step.

        Returns:
            An observation of the system at the current time step.

        Note:
            Override :py:meth:`generate_observation` in subclasses if the system is
            partially observable. By default, the function returns the system state
            directly, meaning it is fully observable.

        """

        return np.array(state, dtype=np.float32)

    def cost(self, time, state, action):
        """Cost function for the system.

        Warning:
            This function is typically not used in SOCKS, but is included here for
            compatibility with OpenAI gym, which returns the cost from the
            :py:meth:`step` function.

        """

        return 0.0

    def step(self, time=0, action=None) -> tuple:
        """Advances the system forward one time step.

        Args:
            time: Time of the simulation. Used primarily for time-varying systems.
            action: Action (input) applied to the system at the current time step.

        Returns:
            A tuple ``(obs, cost, done, info)``, where ``obs`` is the observation
            vector. Generally, it is the state of the system corrupted by some
            measurement noise. If the system is fully observable, this is the actual
            state of the system at the next time step. ``cost`` is the cost
            (reward) obtained by the system for taking action u in state x and
            transitioning to state y. In general, this is not typically used with
            :py:class:`DynamicalSystem` models. ``done`` is a flag to indicate the
            simulation has terminated. Usually toggled by guard conditions, which
            terminates the simulation if the system violates certain operating
            constraints. ``info`` is a dictionary containing extra information.

        """

        action = np.asarray(action, dtype=np.float32)

        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        # Generate disturbance.
        disturbance = self.generate_disturbance(time, self.state, action)

        if self._euler is True:
            next_state = self.state + self.sampling_time * self.dynamics(
                time, self.state, action, disturbance
            )
            self.state = next_state
        else:
            # Solve the initial value problem.
            sol = solve_ivp(
                self.dynamics,
                [time, time + self.sampling_time],
                self.state,
                args=(
                    action,
                    disturbance,
                ),
            )
            *_, self.state = sol.y.T

        # Generate observation.
        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def reset(self):
        """Reset the system to a random initial condition."""
        self.state = self.state_space.sample()
        return np.array(self.state, dtype=np.float32)

    def render(self, mode="human"):
        raise NotImplementedError

    def close(self):
        pass

    @property
    def np_random(self):
        """Random number generator."""
        if self._np_random is None:
            self.seed()

        return self._np_random

    def seed(self, seed=None):
        self._np_random, seed = seeding.np_random(seed)
        return [seed]



import unittest
from unittest.mock import patch

import gym

from gym_socks.envs import NDIntegratorEnv
from gym_socks.envs.core import BaseWrapper, pre_hook_wrapper
from gym_socks.envs.core import post_hook_wrapper

import numpy as np


def custom_pre_hook():
    # print("PRE")
    pass


def custom_post_hook():
    # print("POST")
    pass


class DummyWrapper(NDIntegratorEnv):
    @pre_hook_wrapper(custom_pre_hook)
    @post_hook_wrapper(custom_post_hook)
    def reset(self):
        return


class TestWrapper(unittest.TestCase):
    # @patch.object(HookWrapper, "pre_hook", new=custom_pre_hook)
    # @patch.object(HookWrapper, "post_hook", new=custom_post_hook)
    def test_pre_post_hook(cls):

        # mock_pre_hook.new_callable = custom_pre_hook
        # mock_post_hook.new_callable = custom_post_hook

        # wrapped = HookWrapper(NDIntegratorEnv())
        wrapped = DummyWrapper()

        wrapped.reset()

        pass

    def test_wrapper_class(cls):

        wrapped = BaseWrapper(NDIntegratorEnv(2))

        # print(wrapped.reset())



from abc import abstractmethod

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np

from scipy.constants import gravitational_constant


class BaseCWH(object):
    """CWH base class.

    This class holds the shared parameters for the CWH systems, which include:

    * orbital radius
    * gravitational constant
    * celestial body mass
    * chief mass

    And provides methods to compute:

    * graviational parameter (:math:`\mu`)
    * angular velocity (n)

    """

    time_horizon = 600  # [s]
    sampling_time = 20  # [s]

    # system parameters
    _orbital_radius = 850 + 6378.1  # [m]
    _gravitational_constant = gravitational_constant
    _celestial_body_mass = 5.9472e24  # [kg]
    _chief_mass = 300  # [kg]

    @property
    def orbital_radius(self):
        return self._orbital_radius

    @orbital_radius.setter
    def orbital_radius(self, value):
        self._orbital_radius = value
        self.compute_angular_velocity()

    @property
    def gravitational_constant(self):
        return self._gravitational_constant

    @gravitational_constant.setter
    def gravitational_constant(self, value):
        self._gravitational_constant = value
        self.compute_mu()

    @property
    def celestial_body_mass(self):
        return self._celestial_body_mass

    @celestial_body_mass.setter
    def celestial_body_mass(self, value):
        self._celestial_body_mass = value
        self.compute_mu()

    @property
    def chief_mass(self):
        return self._chief_mass

    @chief_mass.setter
    def chief_mass(self, value):
        self._chief_mass = value

    def compute_mu(self):
        self.mu = self.gravitational_constant * self.celestial_body_mass / 1e9
        self.compute_angular_velocity()

    def compute_angular_velocity(self):
        self.angular_velocity = np.sqrt(self.mu / (self.orbital_radius ** 3))

        self.state_matrix = self.compute_state_matrix(sampling_time=self.sampling_time)
        self.input_matrix = self.compute_input_matrix(sampling_time=self.sampling_time)


class CWH4DEnv(BaseCWH, DynamicalSystem):
    """4D Clohessy-Wiltshire-Hill (CWH) system.

    Bases: :py:class:`gym_socks.envs.cwh.BaseCWH`, :py:class:`gym_socks.envs.dynamical_system.DynamicalSystem`

    The 4D CWH system is a simplification of the 6D dynamics to operate within a plane.
    Essentially, it ignores the :math:`z` component of the dynamics.

    """

    def __init__(self, seed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.observation_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32
        )
        self.state_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32
        )
        self.action_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(2,), dtype=np.float32
        )

        self.state = None

        self.seed(seed=seed)

        self.compute_mu()
        self.compute_angular_velocity()

        self.state_matrix = self.compute_state_matrix(sampling_time=self.sampling_time)
        self.input_matrix = self.compute_input_matrix(sampling_time=self.sampling_time)

    def compute_state_matrix(self, sampling_time):
        n = self.angular_velocity
        nt = n * sampling_time
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        return [
            [4 - 3 * cos_nt, 0, (1 / n) * sin_nt, (2 / n) * (1 - cos_nt)],
            [
                6 * (sin_nt - nt),
                1,
                -(2 / n) * (1 - cos_nt),
                (1 / n) * (4 * sin_nt - 3 * nt),
            ],
            [3 * n * sin_nt, 0, cos_nt, 2 * sin_nt],
            [-6 * n * (1 - cos_nt), 0, -2 * sin_nt, 4 * cos_nt - 3],
        ]

    def compute_input_matrix(self, sampling_time):
        n = self.angular_velocity
        Ts = sampling_time
        nt = n * Ts
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        sin_nt2 = np.sin((nt) / 2)

        M = self.chief_mass

        eAt = [
            [
                4 * Ts - (3 * sin_nt) / n,
                0,
                (2 * (sin_nt2 ** 2)) / (n ** 2),
                -(2 * (sin_nt - nt)) / (n ** 2),
            ],
            [
                (12 * (sin_nt2 ** 2)) / n - 3 * (Ts ** 2) * n,
                Ts,
                (2 * (sin_nt - nt)) / (n ** 2),
                (8 * (sin_nt2 ** 2)) / (n ** 2) - (3 * (Ts ** 2)) / 2,
            ],
            [3 - 3 * cos_nt, 0, sin_nt / n, (4 * (sin_nt2 ** 2)) / n],
            [
                6 * sin_nt - 6 * nt,
                0,
                -(4 * (sin_nt2 ** 2)) / n,
                (4 * sin_nt) / n - 3 * Ts,
            ],
        ]

        B = [
            [0, 0],
            [0, 0],
            [(1 / M), 0],
            [0, (1 / M)],
        ]

        return np.matmul(eAt, B)

    def step(self, action, time=0):
        action = np.asarray(action, dtype=np.float32)

        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        disturbance = self.generate_disturbance(time, self.state, action)
        self.state = self.dynamics(time, self.state, action, disturbance)
        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        w = np.multiply([1e-4, 1e-4, 5e-8, 5e-8], w)
        return np.array(w)

    def dynamics(self, time, state, action, disturbance):
        """Dynamics for the system.

        Attention:
            For this system, the :py:obj:`scipy.solve_ivp` function does not return the
            correct result for the dynamical equations, and will quickly run into
            numerical issues where the states explode. However, the CWH system has a
            closed-form solution for the equations of motion, so we use this instead to
            calculate the evolution of the system. See the :py:meth:`step` function for
            additional details regarding how the next state is calculated.

        """

        return (
            np.matmul(self.state_matrix, state)
            + np.matmul(self.input_matrix, action)
            + disturbance
        )

        # x1, x2, x3, x4 = state
        # u1, u2 = action
        # w1, w2, w3, w4 = disturbance

        # dx1 = x1 + w1
        # dx2 = x2 + w2
        # dx3 = (
        #     3 * (self.angular_velocity ** 2) * x1
        #     + 2 * self.angular_velocity * x4
        #     + (u1 / self.chief_mass)
        #     + w3
        # )
        # dx4 = -2 * self.angular_velocity * x3 + (u2 / self.chief_mass) + w4

        # return np.array([dx1, dx2, dx3, dx4], dtype=np.float32)


class CWH6DEnv(BaseCWH, DynamicalSystem):
    """6D Clohessy-Wiltshire-Hill (CWH) system.

    Bases: :py:class:`gym_socks.envs.dynamical_system.DynamicalSystem`

    """

    def __init__(self, seed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.observation_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32
        )
        self.state_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32
        )
        self.action_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(3,), dtype=np.float32
        )

        self.state = None

        self.seed(seed=seed)

        self.compute_mu()
        self.compute_angular_velocity()

        self.state_matrix = self.compute_state_matrix(sampling_time=self.sampling_time)
        self.input_matrix = self.compute_input_matrix(sampling_time=self.sampling_time)

    def compute_state_matrix(self, sampling_time):
        n = self.angular_velocity
        nt = n * sampling_time
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        return [
            [4 - 3 * cos_nt, 0, 0, (1 / n) * sin_nt, (2 / n) * (1 - cos_nt), 0],
            [
                6 * (sin_nt - nt),
                1,
                0,
                -(2 / n) * (1 - cos_nt),
                (1 / n) * (4 * sin_nt - 3 * nt),
                0,
            ],
            [0, 0, cos_nt, 0, 0, (1 / n) * sin_nt],
            [3 * n * sin_nt, 0, 0, cos_nt, 2 * sin_nt, 0],
            [-6 * n * (1 - cos_nt), 0, 0, -2 * sin_nt, 4 * cos_nt - 3, 0],
            [0, 0, -n * sin_nt, 0, 0, cos_nt],
        ]

    def compute_input_matrix(self, sampling_time):
        n = self.angular_velocity
        Ts = sampling_time
        nt = n * Ts
        sin_nt = np.sin(nt)
        cos_nt = np.cos(nt)

        sin_nt2 = np.sin((nt) / 2)

        M = self.chief_mass

        eAt = [
            [
                4 * Ts - (3 * sin_nt) / n,
                0,
                0,
                (2 * (sin_nt2 ** 2)) / (n ** 2),
                -(2 * (sin_nt - nt)) / (n ** 2),
                0,
            ],
            [
                (12 * (sin_nt2 ** 2)) / n - 3 * (Ts ** 2) * n,
                Ts,
                0,
                (2 * (sin_nt - nt)) / (n ** 2),
                (8 * (sin_nt2 ** 2)) / (n ** 2) - (3 * (Ts ** 2)) / 2,
                0,
            ],
            [0, 0, sin_nt / n, 0, 0, (2 * (sin_nt2 ** 2)) / (n ** 2)],
            [
                3 - 3 * cos_nt,
                0,
                0,
                sin_nt / n,
                (4 * (sin_nt2 ** 2)) / n,
                0,
            ],
            [
                6 * sin_nt - 6 * nt,
                0,
                0,
                -(4 * (sin_nt2 ** 2)) / n,
                (4 * sin_nt) / n - 3 * Ts,
                0,
            ],
            [0, 0, cos_nt - 1, 0, 0, sin_nt / n],
        ]

        B = [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [(1 / M), 0, 0],
            [0, (1 / M), 0],
            [0, 0, (1 / M)],
        ]

        return np.matmul(eAt, B)

    def step(self, action, time=0):
        action = np.asarray(action, dtype=np.float32)

        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg

        disturbance = self.generate_disturbance(time, self.state, action)
        self.state = self.dynamics(time, self.state, action, disturbance)
        observation = self.generate_observation(time, self.state, action)

        cost = self.cost(time, self.state, action)

        done = False
        info = {}

        return observation, cost, done, info

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        w = np.multiply([1e-4, 1e-4, 1e-4, 5e-8, 5e-8, 5e-8], w)
        return np.array(w)

    def dynamics(self, time, state, action, disturbance):
        """Dynamics for the system.

        Attention:
            For this system, the :py:obj:`scipy.solve_ivp` function does not return the
            correct result for the dynamical equations, and will quickly run into
            numerical issues where the states explode. However, the CWH system has a
            closed-form solution for the equations of motion, so we use this instead to
            calculate the evolution of the system. See the :py:meth:`step` function for
            additional details regarding how the next state is calculated.

        """

        return (
            np.matmul(self.state_matrix, self.state)
            + np.matmul(self.input_matrix, action)
            + disturbance
        )

        # x1, x2, x3, x4, x5, x6 = state
        # u1, u2, u3 = action
        # w1, w2, w3, w4, w5, w6 = disturbance

        # dx1 = x1 + w1
        # dx2 = x2 + w2
        # dx3 = x3 + w3
        # dx4 = (
        #     3 * (self.angular_velocity ** 2) * x1
        #     + 2 * self.angular_velocity * x5
        #     + (u1 / self.chief_mass)
        #     + w4
        # )
        # dx5 = -2 * self.angular_velocity * x4 + (u2 / self.chief_mass) + w5
        # dx6 = -(self.angular_velocity ** 2) * x3 + (u3 / self.chief_mass) + w6

        # return np.array([dx1, dx2, dx3, dx4, dx5, dx6], dtype=np.float32)



"""Kernel-based linear system identification.

The algorithm uses a concatenated state space representation to compute the state and
input matrices given a sample of system observations. Uses the matrix inversion lemma
and a linear kernel to compute the linear relationship between observations.

"""

from functools import partial

import numpy as np

from gym_socks.algorithms.base import RegressorMixin
from gym_socks.sampling.transform import transpose_sample

import logging
from gym_socks.utils.logging import ms_tqdm, _progress_fmt

logger = logging.getLogger(__name__)


def kernel_linear_id(
    S: np.ndarray,
    regularization_param: float = None,
    verbose: bool = False,
) -> np.ndarray:
    """Stochastic linear system identification using kernel distribution embeddings.

    Computes an approximation of the state and input matrices, as well as an estimate of
    the stochastic disturbance mean given a sample of system observations.

    Args:
        S: Sample of (x, u, y) tuples taken iid from the system evolution. The sample
            should be in the form of a list of tuples.
        regularization_param: Regularization parameter used in the solution to the
            regularized least-squares problem.
        verbose: Boolean flag to indicate verbose output.

    Returns:
        The fitted model computed using the linear identification algorithm.

    """

    alg = KernelLinearId(
        regularization_param=regularization_param,
        verbose=verbose,
    )
    alg.fit(S)
    return alg


class KernelLinearId(RegressorMixin):
    """Stochastic linear system identification using kernel distribution embeddings.

    Computes an approximation of the state and input matrices, as well as an estimate of
    the stochastic disturbance mean given a sample of system observations.

    Args:
        regularization_param : Regularization parameter used in the solution to the
            regularized least-squares problem.
        verbose : Boolean flag to indicate verbose output.

    """

    def __init__(
        self,
        regularization_param: float = None,
        verbose: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.regularization_param = regularization_param

        self.verbose = verbose

        self._state_matrix = None
        self._input_matrix = None
        self._disturbance_mean = None

    @property
    def state_matrix(self):
        """The estimated state matrix."""
        return self._state_matrix

    @property
    def input_matrix(self):
        """The estimated input matrix."""
        return self._input_matrix

    @property
    def disturbance_mean(self):
        """The estimated stochastic disturbance mean."""
        return self._disturbance_mean

    def _validate_params(self, S):

        if self.regularization_param is None:
            self.regularization_param = 1

    def _validate_data(self, S):

        if S is None:
            raise ValueError("Must supply a sample.")

    def fit(self, S: np.ndarray):
        """Run the algorithm.

        Args:
            S : Sample of (x, u, y) tuples taken iid from the system evolution. The
                sample should be in the form of a list of tuples.

        Returns:
            self : Instance of KernelLinearId class.

        """

        self._validate_data(S)
        self._validate_params(S)

        len_sample = len(S)

        X, U, Y = transpose_sample(S)
        X = np.array(X)
        U = np.array(U)
        Y = np.array(Y)

        x_dim = X.shape[1]  # dimensionality of the input data
        u_dim = U.shape[1]  # dimensionality of the input data

        # Create concatenated state representation.
        Z = np.concatenate((X, U, np.ones((len_sample, 1))), axis=1)

        z_dim = Z.shape[1]

        if len_sample < z_dim:
            msg = f"The sample size is less than the dim of H: {len_sample} < {z_dim}."
            logger.warn(msg)

        logger.debug("Computing covariance matrices.")
        CZZ = (Z.T @ Z) + self.regularization_param * len_sample * np.identity(z_dim)
        CYZ = Y.T @ Z

        logger.debug("Computing estimates.")
        H = np.linalg.solve(
            CZZ.T,
            CYZ.T,
        ).T

        self._state_matrix = H[:x_dim, :x_dim]
        self._input_matrix = H[:x_dim, x_dim : x_dim + u_dim]
        self._disturbance_mean = H[:x_dim, -1:]

        return self

    def predict(self, T: np.ndarray, U: np.ndarray = None) -> np.ndarray:
        """Predict.

        Args:
            T : State vectors. Should be in the form of a 2D-array, where each row
                indicates a state.
            U : Input vectors. Should be in the form of a 2D-array, where each row
                indicates an input.

        Returns:
            The predicted resulting states after propagating through the estimated
            system dynamics.

        """

        self._validate_data(T)
        T = np.array(T)

        result = self._state_matrix @ T.T

        if U is not None:
            self._validate_data(U)
            U = np.array(U)

            result += self._input_matrix @ U.T

        result += np.squeeze(self._disturbance_mean.T)

        return result

    def score(self):
        raise NotImplementedError



import unittest

import gym

import gym_socks.utils

import numpy as np


class TestNormalize(unittest.TestCase):
    def test_normalize(cls):
        """Test normalize function."""

        # Single point.
        points = [1.0, 2.0]
        groundTruth = [0.33333333, 0.66666667]
        normalize_result = gym_socks.utils.normalize(points)
        cls.assertTrue(np.allclose(normalize_result, groundTruth))

        # Multiple points.
        points = [[1.0, 2.0], [0.5, 1.0]]
        groundTruth = [[0.66666667, 0.66666667], [0.33333333, 0.33333333]]
        normalize_result = gym_socks.utils.normalize(points)
        cls.assertTrue(np.allclose(normalize_result, groundTruth))

        points = [[1.0, 0.5], [2.0, 1.0]]
        groundTruth = [[0.33333333, 0.33333333], [0.66666667, 0.66666667]]
        normalize_result = gym_socks.utils.normalize(points)
        cls.assertTrue(np.allclose(normalize_result, groundTruth))


class TestIndicatorFunction(unittest.TestCase):
    def test_indicator_function_on_box(cls):
        """Test indicator function on box."""

        interval = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        # One point inside.
        points = [[0.1, 0.1]]
        groundTruth = np.array([True], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, space=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))

        # Multiple points inside.
        points = [[0.1, 0.1], [0.2, -0.2]]
        groundTruth = np.array([True, True], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, space=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))

        # One point in, one point out.
        points = [[0.1, 0.1], [1.2, -0.2]]
        groundTruth = np.array([True, False], dtype=bool)
        indicator_result = gym_socks.utils.indicator_fn(points=points, space=interval)
        cls.assertTrue(np.all(np.equal(indicator_result, groundTruth)))



import numpy as np

from functools import partial

from gym_socks.kernel.metrics import regularized_inverse
from gym_socks.kernel.metrics import rbf_kernel


def maximum_mean_discrepancy(
    X, Y, kernel_fn=None, biased: bool = False, squared: bool = False
):
    r"""Maximum mean discrepancy between two empirical distributions.

    Compute the maximum mean discrepancy between two distributions (samples :math:`X`
    and :math:`Y`) using a kernel-based statistic. The biased estimator
    :math:`\widehat{\textrm{MMD}}_{b}^{2}(X, Y, \mathscr{H})` uses U-statistics, but can
    be negative. The unbiased estimator :math:`\widehat{\textrm{MMD}}_{u}^{2}(X, Y,
    \mathscr{H})` uses V-statistics.

    .. math::

        \widehat{\textrm{MMD}}_{b}^{2}(X, Y, \mathscr{H}) =
        \frac{1}{m^{2}} \sum_{i=1}^{m} \sum_{j=1}^{m} k(x_{i}, x_{j})
        + \frac{1}{n^{2}} \sum_{i=1}^{n} \sum_{j=1}^{n} k(y_{i}, y_{j})
        - \frac{2}{mn} \sum_{i=1}^{m} \sum_{j=1}^{n} k(x_{i}, y_{j})

    .. math::

        \widehat{\textrm{MMD}}_{u}^{2}(X, Y, \mathscr{H}) =
        \frac{1}{m(m - 1)} \sum_{i=1}^{m} \sum_{j \neq i}^{m} k(x_{i}, x_{j})
        + \frac{1}{n(n - 1)} \sum_{i=1}^{n} \sum_{j \neq i}^{n} k(y_{i}, y_{j})
        - \frac{2}{mn} \sum_{i=1}^{m} \sum_{j=1}^{n} k(x_{i}, y_{j})

    Args:
        X: The observations from distribution P organized in ROWS.
        Y: The observations from distribution Q organized in ROWS.
        kernel_fn: The kernel function is a function that returns an ndarray where
            each element is the pairwise evaluation of a kernel function. See sklearn.
            metrics.pairwise for more info. The default is the RBF kernel.
        biased: Whether to use the biased MMD.
        squared: Whether or not the result is squared.

    Returns:
        The scalar value representing the MMD.

    """

    m = len(X)
    n = len(Y)

    if kernel_fn is None:
        kernel_fn = partial(rbf_kernel, sigma=1)

    GX = kernel_fn(X)
    GY = kernel_fn(Y)
    GXY = kernel_fn(X, Y)

    if biased is True:
        # Biased statistic.
        c1 = 1 / (m ** 2)
        c2 = 1 / (n ** 2)
        c3 = 2 / (m * n)
        A = np.sum(GX)
        B = np.sum(GY)
        C = np.sum(GXY)
        mmd = c1 * A + c2 * B - c3 * C
    else:
        # Unbiased statistic.
        c1 = 1 / (m * (m - 1))
        c2 = 1 / (n * (n - 1))
        c3 = 2 / (m * n)
        A = np.sum(GX - np.diag(np.diag(GX)))
        B = np.sum(GY - np.diag(np.diag(GY)))
        C = np.sum(GXY)
        mmd = c1 * A + c2 * B - c3 * C

    return mmd if squared else np.sqrt(mmd)


def witness_function(X, Y, t, kernel_fn=None):

    m = len(X)
    n = len(Y)

    if kernel_fn is None:
        kernel_fn = partial(rbf_kernel, sigma=1)

    c1 = 1 / m
    c2 = 1 / n
    A = np.sum(kernel_fn(X, t), axis=0)
    B = np.sum(kernel_fn(Y, t), axis=0)

    return c1 * A - c2 * B


def kernel_sum_rule(
    G,
    K,
    alpha,
    regularization_param: float = None,
):
    """Computes the kernel sum rule using Gram matrices.

    The kernel sum rule computes the marginal of a conditional distribution by
    integrating out the conditioning variables.

    .. math::

        P(Y) = \int_{\Omega} P(Y \mid X) P(d X)

    Args:
        G: The gram matrix of Y. Typically computed as `G = kernel_fn(Y)`.
        K: The Gram matrix of X. Typically computed as `K = kernel_fn(X)`.
        alpha: The coefficient vector corresponding to the marginal embedding of P(X).
        regularization_param: The regularization parameter of the matrix inverse.

    Returns:
        The coefficients of the marginal embedding of P(Y).

    """

    num_rows, _ = G.shape

    return np.einsum(
        "ii,ij,j->j",
        np.linalg.inv(G + regularization_param * num_rows * np.identity(num_rows)),
        K,
        alpha,
    )


def kernel_chain_rule(
    G,
    K,
    alpha,
    regularization_param: float = None,
):
    """Computes the kernel chain rule using Gram matrices.

    The kernel chain rule computes the joint distribution embedding by taking the
    product of conditional and marginal distribution embeddings.

    .. math::

        P(Y, X) = P(Y \mid X) P(X)

    Args:
        G: The gram matrix of Y. Typically computed as `G = kernel_fn(Y)`.
        K: The Gram matrix of X. Typically computed as `K = kernel_fn(X)`.
        alpha: The coefficient vector corresponding to the marginal embedding of P(X).
        regularization_param: The regularization parameter of the matrix inverse.

    Returns:
        The coefficients of the joint distribution embedding of P(Y, X).

    """

    num_rows, _ = G.shape

    return np.einsum(
        "ii,ij,jk->ik",
        np.linalg.inv(G + regularization_param * num_rows * np.identity(num_rows)),
        K,
        np.diag(alpha),
    )


def kernel_bayes_rule(
    G,
    K,
    Kx,
    alpha,
    regularization_param: float = None,
):
    r"""Computes the kernel Bayes' rule using Gram matrices.

    The kernel Bayes' rule.

    .. math::

        P(Y \mid x) = \frac{P(x \mid Y) P(Y)}{P(x)},
        \quad P(x) = \int_{\Omega} P(x \mid y) P(d y)

    Args:
        G: The gram matrix of Y. Typically computed as `G = kernel_fn(Y)`.
        K: The Gram matrix of X. Typically computed as `K = kernel_fn(X)`.
        Kx: The feature vector of x. Typically computed as `Kx = kernel_fn(X, x)`.
        alpha: The coefficient vector corresponding to the marginal embedding of P(X).
        regularization_param: The regularization parameter of the matrix inverse.

    Returns:
        The coefficients of the conditional distribution embedding of P(Y | x).

    """

    num_rows, _ = G.shape

    W = np.linalg.inv(G + regularization_param * num_rows * np.identity(num_rows))

    L = kernel_chain_rule(G, K, alpha, regularization_param=regularization_param)
    D = np.diag(kernel_sum_rule(G, K, alpha, regularization_param=regularization_param))

    return np.einsum(
        "ji,ii,jk,kl,l->jl",
        L,
        np.linalg.ing(
            np.linalg.matrix_power(D @ K, 2)
            + regularization_param * num_rows * np.identity(num_rows)
        ),
        K,
        D,
        Kx,
    )



import unittest
from unittest import mock
from unittest.mock import patch

import gym

import numpy as np

from gym_socks.algorithms.identification.kernel_linear_id import KernelLinearId
from gym_socks.envs import CWH4DEnv
from gym_socks.envs import CWH6DEnv
from gym_socks.policies import RandomizedPolicy

from gym_socks.sampling import sample
from gym_socks.sampling import sample_generator


def dummy_sampler(env, policy, sample_space):
    @sample_generator
    def _sample_generator():
        state = sample_space.sample()
        action = policy(time=0, state=state)

        env.state = state
        next_state, *_ = env.step(action=action)

        yield (state, action, next_state)

    return _sample_generator


class TestKernelLinearIdentificationAlgorithm(unittest.TestCase):
    def test_linear_identification_CWH4D(cls):
        """Should identify linear dynamics of CWH4D system within a tolerance."""

        env = CWH4DEnv()
        env.seed(seed=0)

        S = sample(
            sampler=dummy_sampler(
                env=env,
                policy=RandomizedPolicy(action_space=env.action_space),
                sample_space=env.state_space,
            ),
            sample_size=1000,
        )

        regularization_param = 1e-9

        alg = KernelLinearId(regularization_param=regularization_param, verbose=False)
        alg.fit(S=S)

        cls.assertTrue(np.allclose(alg.state_matrix, env.state_matrix, atol=1e-2))
        cls.assertTrue(np.allclose(alg.input_matrix, env.input_matrix, atol=1e-2))

    def test_linear_identification_CWH6D(cls):
        """Should identify linear dynamics of CWH6D system within a tolerance."""

        env = CWH6DEnv()
        env.seed(seed=0)

        S = sample(
            sampler=dummy_sampler(
                env=env,
                policy=RandomizedPolicy(action_space=env.action_space),
                sample_space=env.state_space,
            ),
            sample_size=1000,
        )

        regularization_param = 1e-9

        alg = KernelLinearId(regularization_param=regularization_param, verbose=False)
        alg.fit(S=S)

        cls.assertTrue(np.allclose(alg.state_matrix, env.state_matrix, atol=1e-2))
        cls.assertTrue(np.allclose(alg.input_matrix, env.input_matrix, atol=1e-2))



import unittest
from unittest.mock import Base, patch

import gym

from gym_socks.envs import NDIntegratorEnv
from gym_socks.envs.core import BaseDynamicalObject

from gym_socks.policies import RandomizedPolicy

from gym_socks.envs.obstacle import BaseObstacle
from gym_socks.envs.world import World

import numpy as np


class DummyObstacle(BaseObstacle):
    def __init__(self, identifier: str) -> None:
        self._id = identifier

    def intersects(self, state):
        return

    def step(self, time=0):
        # print(f"Hello, {self._id}.")
        pass

    def reset(self):
        pass

    def render(self):
        raise NotImplementedError


class TestWorld(unittest.TestCase):
    def test_world(cls):
        """Test world iteration."""

        env = NDIntegratorEnv()
        policy = RandomizedPolicy(action_space=env.action_space)

        obstacle1 = DummyObstacle(identifier="Larry")
        obstacle2 = DummyObstacle(identifier="Curly")
        obstacle3 = DummyObstacle(identifier="Moe")

        world = World()
        world.time_horizon = 5

        world += [obstacle1, obstacle2, obstacle3]

        world.reset()
        env.reset()

        for t in range(world.time_horizon):
            world.step()

            action = policy(time=t, state=[env.state])
            env.step(time=t, action=action)



import unittest
from unittest import mock
from unittest.mock import patch

import gym
import gym_socks

from scipy.constants.codata import unit

from gym_socks.envs.integrator import NDIntegratorEnv
from gym_socks.envs.dynamical_system import DynamicalSystem
from gym_socks.policies import ZeroPolicy
from gym_socks.policies import RandomizedPolicy

from gym_socks.sampling import sample
from gym_socks.sampling import default_trajectory_sampler
from gym_socks.sampling import random_sampler
from gym_socks.sampling import sample_generator
from gym_socks.sampling.transform import transpose_sample

from gym_socks.algorithms.reach.common import _tht_step, _fht_step
from gym_socks.algorithms.reach.monte_carlo import MonteCarloSR
from gym_socks.algorithms.reach.monte_carlo import monte_carlo_sr
from gym_socks.algorithms.reach.monte_carlo import _trajectory_indicator

import numpy as np


def make_tube(
    time_horizon: int, shape: tuple, lower_bound: float, upper_bound: float
) -> list:
    """Creates a tube for testing."""

    tube = []
    for i in range(time_horizon):
        tube_t = gym.spaces.Box(
            lower_bound,
            upper_bound,
            shape=shape,
            dtype=np.float32,
        )
        tube.append(tube_t)

    return tube


@sample_generator
def zero_sampler(sample_space):
    yield np.zeros(shape=sample_space.shape, dtype=sample_space.dtype)


class TestTrajectoryIndicator(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.env = NDIntegratorEnv(2)

        cls.time_horizon = 20

        cls.target_tube = make_tube(
            time_horizon=cls.time_horizon + 1,
            shape=cls.env.state_space.shape,
            lower_bound=-0.25,
            upper_bound=0.25,
        )
        cls.constraint_tube = make_tube(
            time_horizon=cls.time_horizon + 1,
            shape=cls.env.state_space.shape,
            lower_bound=-1,
            upper_bound=1,
        )

    @patch.object(gym_socks.envs.NDIntegratorEnv, "generate_disturbance")
    def test_trajectory_indicator_inside(cls, mock_generate_disturbance):
        """Indicator should return one if trajectory satisfies constraints."""
        mock_generate_disturbance.return_value = np.zeros((2,))

        env = NDIntegratorEnv(2)

        # Test inside.

        # The trajectory of the system with no disturbance starts at (-0.1, 0.1) and
        # ends at (0.1, 0.1) after 20 time steps.

        sample_space = gym.spaces.Box(
            low=np.array([-0.1, 0.1], dtype=np.float32),
            high=np.array([-0.1, 0.1], dtype=np.float32),
            dtype=np.float32,
        )

        sample_size = 5

        S = sample(
            sampler=default_trajectory_sampler(
                state_sampler=random_sampler(sample_space=sample_space),
                action_sampler=zero_sampler(sample_space=env.action_space),
                env=env,
                time_horizon=cls.time_horizon,
            ),
            sample_size=sample_size,
        )

        X, _, Y = transpose_sample(S)
        full_trajectories = [(X[i], *Y[i]) for i in range(sample_size)]

        result = _trajectory_indicator(
            trajectories=full_trajectories,
            time_horizon=cls.time_horizon,
            constraint_tube=cls.constraint_tube,
            target_tube=cls.target_tube,
            step_fn=_tht_step,
        )

        cls.assertTrue(np.all(result))

    @patch.object(gym_socks.envs.NDIntegratorEnv, "generate_disturbance")
    def test_trajectory_indicator_outside(cls, mock_generate_disturbance):
        """Indicator should return zero if trajectory does not satisfy constraints."""
        mock_generate_disturbance.return_value = np.zeros((2,))

        env = NDIntegratorEnv(2)

        # Test not inside.

        # The trajectory of the system with no disturbance starts at (0.2, 0.1) and
        # ends at (0.4, 0.1) after 20 time steps.

        sample_space = gym.spaces.Box(
            low=np.array([0.2, 0.1], dtype=np.float32),
            high=np.array([0.2, 0.1], dtype=np.float32),
            dtype=np.float32,
        )

        sample_size = 5

        S = sample(
            sampler=default_trajectory_sampler(
                state_sampler=random_sampler(sample_space=sample_space),
                action_sampler=zero_sampler(sample_space=env.action_space),
                env=env,
                time_horizon=cls.time_horizon,
            ),
            sample_size=sample_size,
        )

        X, _, Y = transpose_sample(S)
        full_trajectories = [(X[i], *Y[i]) for i in range(sample_size)]

        result = _trajectory_indicator(
            trajectories=full_trajectories,
            time_horizon=cls.time_horizon,
            constraint_tube=cls.constraint_tube,
            target_tube=cls.target_tube,
            step_fn=_tht_step,
        )

        cls.assertTrue(~np.any(result))



import numpy as np


def transpose_sample(sample):
    """Transpose the sample.

    By default, a sample should be a list of tuples of the form::

        S = [(x_1, y_1), ..., (x_n, y_n)]

    For most algorithms, we need to isolate the sample components (e.g. all x's).
    This function converts a sample from a list of tuples to a tuple of lists::

        S_T = ([x_1, ..., x_n], [y_1, ..., y_n])

    This can then be unpacked as: ``X, Y = S_T``

    Args:
        sample: list of tuples

    Returns:
        tuple of lists

    """

    return tuple(map(list, zip(*sample)))


def flatten_sample(sample):
    """Reshapes trajectory samples.

    Often, trajectory samples are organized such that the "trajectory" components are a
    2D array of points indexed by time. However, for kernel methods, we typically
    require that the trajectories be concatenated into a single vector (1D array)::

        [[x1], [x2], ..., [xn]] -> [x1, x2, ..., xn]

    This function converts the sample so that the trajectories are 1D arrays.

    Args:
        sample: list of tuples

    Returns:
        List of tuples, where the components of the tuples are flattened.

    """
    sample_size = len(sample)
    _S = transpose_sample(sample)

    return zip(*[np.reshape(item, (sample_size, -1)) for item in _S])



import unittest
from unittest.mock import patch

import gym

from gym_socks.envs import NDIntegratorEnv
from gym_socks.policies import ConstantPolicy
from gym_socks.policies import RandomizedPolicy

from gym_socks.sampling.sample import sample_generator
from gym_socks.sampling.sample import sample
from gym_socks.sampling.sample import default_sampler
from gym_socks.sampling.sample import default_trajectory_sampler
from gym_socks.sampling.sample import random_sampler
from gym_socks.sampling.sample import grid_sampler
from gym_socks.sampling.sample import repeat

from gym_socks.sampling.transform import transpose_sample
from gym_socks.sampling.transform import flatten_sample

from gym_socks.utils.grid import make_grid_from_ranges
from gym_socks.utils.grid import make_grid_from_space

import numpy as np


class TestSample(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.env = NDIntegratorEnv(2)
        cls.policy = RandomizedPolicy(action_space=cls.env.action_space)

        cls.sample_space = gym.spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        cls.sample_space.seed(1)

    def test_random_sampler(cls):
        """Random sampler should output correct sample size."""

        sample_size = 10

        state_sampler = random_sampler(sample_space=cls.env.state_space)
        action_sampler = random_sampler(sample_space=cls.env.action_space)

        S = sample(
            sampler=default_sampler(
                state_sampler=state_sampler,
                action_sampler=action_sampler,
                env=cls.env,
            ),
            sample_size=sample_size,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.shape(X), (sample_size, 2))
        cls.assertEqual(np.shape(U), (sample_size, 1))
        cls.assertEqual(np.shape(Y), (sample_size, 2))

    def test_grid_sampler(cls):
        """Grid sampler should output items on a grid of the correct size."""

        # print(np.repeat(np.linspace(-1, 1, 5), 2, axis=0))

        sample_size = 27

        state_sampler = repeat(
            grid_sampler(
                make_grid_from_space(
                    sample_space=gym.spaces.Box(
                        low=-1,
                        high=1,
                        shape=cls.env.state_space.shape,
                        dtype=cls.env.state_space.dtype,
                    ),
                    resolution=3,
                )
            ),
            num=3,
        )

        action_sampler = grid_sampler(make_grid_from_ranges([np.linspace(-1, 1, 3)]))

        S = sample(
            sampler=default_sampler(
                state_sampler=state_sampler,
                action_sampler=action_sampler,
                env=cls.env,
            ),
            sample_size=sample_size,
        )

        X, U, Y = transpose_sample(S)

        groundTruth = [
            [-1.0, -1.0],
            [-1.0, -1.0],
            [-1.0, -1.0],
            [-1.0, 0.0],
            [-1.0, 0.0],
            [-1.0, 0.0],
            [-1.0, 1.0],
            [-1.0, 1.0],
            [-1.0, 1.0],
            [0.0, -1.0],
            [0.0, -1.0],
            [0.0, -1.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0],
            [0.0, 1.0],
            [0.0, 1.0],
            [1.0, -1.0],
            [1.0, -1.0],
            [1.0, -1.0],
            [1.0, 0.0],
            [1.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [1.0, 1.0],
            [1.0, 1.0],
        ]

        cls.assertTrue(np.array_equiv(X, groundTruth))

        groundTruth = [
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
            [-1.0],
            [0.0],
            [1.0],
        ]

        cls.assertTrue(np.array_equiv(U, groundTruth))

        cls.assertEqual(np.shape(X), (sample_size, 2))
        cls.assertEqual(np.shape(U), (sample_size, 1))
        cls.assertEqual(np.shape(Y), (sample_size, 2))

    def test_default_sampler(cls):
        """Default sampler should output sample of the correct size."""

        sample_size = 10

        state_sampler = random_sampler(sample_space=cls.env.state_space)
        action_sampler = random_sampler(sample_space=cls.env.action_space)

        S = sample(
            sampler=default_sampler(
                state_sampler=state_sampler,
                action_sampler=action_sampler,
                env=cls.env,
            ),
            sample_size=sample_size,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (sample_size, 2))
        cls.assertEqual(np.array(U).shape, (sample_size, 1))
        cls.assertEqual(np.array(Y).shape, (sample_size, 2))

    def test_trajectory_sampler(cls):
        """Default trajectory sampler should output sample of the correct size."""

        sample_size = 10
        time_horizon = 5

        state_sampler = random_sampler(sample_space=cls.env.state_space)
        action_sampler = random_sampler(sample_space=cls.env.action_space)

        S = sample(
            sampler=default_trajectory_sampler(
                state_sampler=state_sampler,
                action_sampler=action_sampler,
                env=cls.env,
                time_horizon=5,
            ),
            sample_size=sample_size,
        )

        cls.assertEqual(np.array(S[0][0]).shape, (2,))
        cls.assertEqual(np.array(S[0][1]).shape, (5, 1))
        cls.assertEqual(np.array(S[0][2]).shape, (5, 2))

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (sample_size, 2))
        cls.assertEqual(np.array(U).shape, (sample_size, 5, 1))
        cls.assertEqual(np.array(Y).shape, (sample_size, 5, 2))

        ST = flatten_sample(S)
        X, U, Y = transpose_sample(ST)

        cls.assertEqual(np.array(X).shape, (sample_size, 2))
        cls.assertEqual(np.array(U).shape, (sample_size, 5))
        cls.assertEqual(np.array(Y).shape, (sample_size, 10))

    def test_custom_sampler_as_function(cls):

        sample_size = 10

        @sample_generator
        def custom_sampler():
            state = cls.sample_space.sample()
            action = cls.env.action_space.sample()

            cls.env.state = state
            next_state, *_ = cls.env.step(action=action)

            return (state, action, next_state)

        #   ^^^^^^

        S = sample(
            sampler=custom_sampler,
            sample_size=sample_size,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (sample_size, 2))
        cls.assertEqual(np.array(U).shape, (sample_size, 1))
        cls.assertEqual(np.array(Y).shape, (sample_size, 2))

    def test_custom_sampler_as_generator(cls):

        sample_size = 10

        @sample_generator
        def custom_sampler():
            state = cls.sample_space.sample()
            action = cls.env.action_space.sample()

            cls.env.state = state
            next_state, *_ = cls.env.step(action=action)

            yield (state, action, next_state)

        #   ^^^^^

        S = sample(
            sampler=custom_sampler,
            sample_size=sample_size,
        )

        X, U, Y = transpose_sample(S)

        cls.assertEqual(np.array(X).shape, (sample_size, 2))
        cls.assertEqual(np.array(U).shape, (sample_size, 1))
        cls.assertEqual(np.array(Y).shape, (sample_size, 2))

    def test_custom_policy_sampler(cls):

        sample_size = 10

        state_sampler = random_sampler(sample_space=cls.env.state_space)
        policy = ConstantPolicy(action_space=cls.env.action_space, constant=-1.0)

        @sample_generator
        def custom_sampler():
            state = next(state_sampler)
            action = policy()

            cls.env.state = state
            next_state, *_ = cls.env.step(action=action)

            yield (state, action, next_state)

        S = sample(
            sampler=custom_sampler,
            sample_size=sample_size,
        )

        X, U, Y = transpose_sample(S)

        cls.assertTrue(np.array_equiv(U, [-1.0] * sample_size))

        cls.assertEqual(np.array(X).shape, (sample_size, 2))
        cls.assertEqual(np.array(U).shape, (sample_size, 1))
        cls.assertEqual(np.array(Y).shape, (sample_size, 2))



from abc import ABC, abstractmethod

import gym
from gym.utils import seeding

import numpy as np

from functools import wraps


class BaseDynamicalObject(gym.Env, ABC):
    """Base dynamical object class.

    Bases: :py:obj:`gym.Env`, :py:obj:`abc.ABC`

    This class is **abstract**, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from :py:class:`BaseDynamicalObject`.

    This class serves as the base interface for dynamical objects, represented in most
    cases by either a :py:class:`~gym_socks.envs.dynamical_system.DynamicalSystem` or an
    obstacle.

    """

    @abstractmethod
    def step(self):
        """Advance the object forward in time.

        Advances the object in the simulation. By default, an object is uncontrolled,
        meaning it accepts no parameters and the system evolves forward in time
        according to its own, internal dynamics. Controlled systems should accept an
        ``action`` parameter, which represents the user-selected control input.

        Additionally, time-varying systems should also accept a ``time`` parameter.

        Returns:
            A tuple ``(obs, cost, done, info)``, where ``obs`` is the observation
            vector. Generally, it is the state of the system corrupted by some
            measurement noise. If the system is fully observable, this is the actual
            state of the system at the next time step. ``cost`` is the cost
            (reward) obtained by the system for taking action u in state x and
            transitioning to state y. In general, this is not typically used with
            :py:class:`DynamicalSystem` models. ``done`` is a flag to indicate the
            simulation has terminated. Usually toggled by guard conditions, which
            terminates the simulation if the system violates certain operating
            constraints. ``info`` is a dictionary containing extra information.

        See also:
            :py:meth:`gym_socks.envs.dynamical_system.DynamicalSystem.step`

        """

        raise NotImplementedError

    @abstractmethod
    def reset(self):
        """Reset the object to an initial state.

        Resets the object to an initial state (which may be random).

        Returns:
            The new initial state of the object.

        """

        raise NotImplementedError

    @abstractmethod
    def render(self, mode="human"):
        """Renders the environment.

        This method must be overridden in subclasses in order to enable rendering.
        Not all environments support rendering.

        Args:
            mode: the mode to render with

        """
        raise NotImplementedError

    def close(self):
        """Override close in your subclass to perform any necessary cleanup.

        Environments will automatically :py:func:`close` themselves when garbage collected or when the program exits.

        """
        pass

    def seed(self, seed=None):
        """Sets the seed of the random number generator.

        Args:
            seed: Integer value representing the random seed.

        Returns:
            The seed of the RNG.

        """

        return


def pre_hook_wrapper(pre_hook):
    def _outer_wrapper(fun):
        @wraps(fun)
        def _wrapper(*args, **kwargs):
            pre_hook()
            result = fun(*args, **kwargs)
            return result

        return _wrapper

    return _outer_wrapper


def post_hook_wrapper(post_hook):
    def _outer_wrapper(fun):
        @wraps(fun)
        def _wrapper(*args, **kwargs):
            result = fun(*args, **kwargs)
            post_hook()
            return result

        return _wrapper

    return _outer_wrapper


class BaseWrapper(object):
    """Wrapper for BaseDynamicalObjects."""

    def __init__(self, env):
        self.__class__ = type(
            env.__class__.__name__, (self.__class__, env.__class__), {}
        )

        self.__dict__ = env.__dict__

        self.env = env


# class HookWrapper(BaseWrapper):
#     def __getattr__(self, attr):
#         """Wrapped __getattr__."""
#         _attr = self.env.__getattribute__(attr)

#         if callable(_attr):

#             @wraps(callable)
#             def _wrapper(*args, **kwargs):
#                 self.pre_hook()
#                 result = _attr(*args, **kwargs)
#                 self.post_hook()

#                 # prevent unwrapping
#                 if result is self.env:
#                     return self

#                 return result

#             return _wrapper

#         else:
#             return _attr

#     # @abstractmethod
#     def pre_hook(self):
#         raise NotImplementedError

#     # @abstractmethod
#     def post_hook(self):
#         raise NotImplementedError

#     # def reset(self):
#     #     self.pre_hook()
#     #     result = self.env.reset()
#     #     self.post_hook()
#     #     return result



__all__ = ["logging", "normalize", "indicator_fn", "generate_batches"]

import gym

import numpy as np


def normalize(v: np.ndarray) -> np.ndarray:
    """Normalize.

    Small utility function for normalizing a matrix or a vector. Divides the matrix
    (vector) by the sum of the matrix rows (vector). Used primarily by the stochastic
    reachability algorithms.

    Args:
        v: Matrix or vector to normalize.

    Returns:
        Normalized matrix or vector.

    """
    return v / np.sum(v, axis=0)


def indicator_fn(points: np.ndarray, space: any) -> np.ndarray:
    """Lightweight indicator for gym.spaces.Box sets.

    Provides a lightweight proxy for the `contains` function in `gym.spaces.Box`. The
    original function does type checking, casting, and shape checking, which is slow.

    Args:
        points: Points to evaluate the indicator at.
        space: Space over which the indicator function is defined. Should be either a
            `gym.spaces.Box` or a function which returns 0 or 1 depending on whether the
            points are inside or outside the space.

    Returns:
        Boolean labels of whether the points are inside or outside the space.

    """

    _l = space.low
    _h = space.high

    return np.array(
        np.all(points >= _l, axis=1) & np.all(points <= _h, axis=1), dtype=bool
    )


def save_mat_file(filename: str, data: dict):
    from scipy.io import savemat

    with open(filename, "wb") as f:
        savemat(f, data)



r"""ND Integrator system.

An integrator system is an extremely simple dynamical system model, typically used
to model a single variable and its higher order derivatives, where the input is
applied to the highest derivative term, and is "integrated" upwards.

.. tab-set::

    .. tab-item:: Continuous Time

        .. math::

            \dot{x} =
            \begin{bmatrix}
                0 & 1 & 0 & \cdots & 0 & 0 \\
                0 & 0 & 1 & \cdots & 0 & 0 \\
                0 & 0 & 0 & \cdots & 0 & 0 \\
                \vdots & \vdots & \vdots & \ddots & \vdots & \vdots \\
                0 & 0 & 0 & \cdots & 0 & 1 \\
                0 & 0 & 0 & \cdots & 0 & 0
            \end{bmatrix} x +
            \begin{bmatrix}
                0 \\
                0 \\
                0 \\
                \vdots \\
                0 \\
                1
            \end{bmatrix} u + w

    .. tab-item:: Discrete Time

        .. math::

            x_{t+1} =
            \begin{bmatrix}
                1 & T & \cdots & \frac{T^{N-2}}{(N-2)!} & \frac{T^{N-1}}{(N-1)!} \\
                0 & 1 & \cdots & \frac{T^{N-3}}{(N-3)!} & \frac{T^{N-2}}{(N-2)!} \\
                \vdots & \vdots & \ddots & \vdots & \vdots \\
                0 & 0 & \cdots & 1 & T \\
                0 & 0 & \cdots & 0 & 1
            \end{bmatrix} x_{t} +
            \begin{bmatrix}
                \frac{T^{N}}{N!} \\
                \frac{T^{N-1}}{(N-1)!} \\
                \vdots \\
                \frac{T^{2}}{2} \\
                T
            \end{bmatrix} u_{t} + w_{t}

A 2D Integrator system, for example, corresponds to the position and velocity
components of a variable, where the input is applied to the velocity and then
integrates upward to the position variable.

Tip:
    Chaining two 2D integrator systems can model a system with x/y position and
    velocity.

    .. tab-set::

        .. tab-item:: Continuous Time

            .. math::

                \dot{x} =
                \begin{bmatrix}
                    0 & 1 & 0 & 0 \\
                    0 & 0 & 0 & 0 \\
                    0 & 0 & 0 & 1 \\
                    0 & 0 & 0 & 0
                \end{bmatrix} x +
                \begin{bmatrix}
                    0 & 0 \\
                    1 & 0 \\
                    0 & 0 \\
                    0 & 1
                \end{bmatrix} u + w

        .. tab-item:: Discrete Time

            .. math::

                x_{t+1} =
                \begin{bmatrix}
                    1 & T & 0 & 0 \\
                    0 & 1 & 0 & 0 \\
                    0 & 0 & 1 & T \\
                    0 & 0 & 0 & 1
                \end{bmatrix} x_{t} +
                \begin{bmatrix}
                    \frac{T^{2}}{2} & 0 \\
                    T & 0 \\
                    0 & \frac{T^{2}}{2} \\
                    0 & T
                \end{bmatrix} u_{t} + w_{t}

"""

import gym
from gym_socks.envs.dynamical_system import DynamicalSystem

import numpy as np


class NDIntegratorEnv(DynamicalSystem):
    """ND integrator system.

    Bases: :py:class:`gym_socks.envs.dynamical_system.DynamicalSystem`

    Args:
        dim: The dimension of the integrator system.

    Example:

        >>> from gym_socks.envs import NDIntegratorEnv
        >>> env = NDIntegratorEnv(dim=2)
        >>> env.reset()
        >>> for i in range(10):
        ...     action = env.action_space.sample()
        ...     obs, reward, done, _ = env.step(action)
        ...
        >>> env.close()

    """

    def __init__(self, dim: int = 2, seed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.observation_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(dim,), dtype=np.float32
        )
        self.state_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(dim,), dtype=np.float32
        )
        self.action_space = gym.spaces.Box(
            low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32
        )

        self.state = None

        self.seed(seed=seed)

    def generate_disturbance(self, time, state, action):
        w = self.np_random.standard_normal(size=self.state_space.shape)
        return 1e-2 * np.array(w)

    def dynamics(self, time, state, action, disturbance):
        _, *x = state
        return np.array([*x, *action], dtype=np.float32) + disturbance



"""Stochastic reachability using Monte-Carlo."""

import numpy as np

from gym_socks.algorithms.base import RegressorMixin
from gym_socks.algorithms.reach.common import _tht_step, _fht_step

from gym_socks.envs.dynamical_system import DynamicalSystem
from gym_socks.policies import BasePolicy

from gym_socks.sampling import sample
from gym_socks.sampling import sample_generator

from gym_socks.utils import indicator_fn

import logging
from gym_socks.utils.logging import ms_tqdm, _progress_fmt

logger = logging.getLogger(__name__)


def _trajectory_indicator(
    trajectories,
    time_horizon=None,
    constraint_tube=None,
    target_tube=None,
    step_fn=None,
):
    trajectories = np.asarray(trajectories, dtype=np.float32)
    result = indicator_fn(trajectories[:, time_horizon], target_tube[time_horizon])
    result = np.array(result, dtype=int)

    for t in range(time_horizon - 1, -1, -1):
        result = step_fn(trajectories[:, t], result, constraint_tube[t], target_tube[t])

    return result


def _monte_carlo_trajectory_sampler(
    time_horizon: int = None,
    env: DynamicalSystem = None,
    policy: BasePolicy = None,
    state: np.ndarray = None,
):
    """Monte-Carlo trajectory sampler.

    Args:
        env: The system to sample from.
        policy: The policy applied to the system during sampling.
        sample_space: The space where initial conditions are drawn from.

    Returns:
        A generator function that yields system observations as tuples.

    """

    @sample_generator
    def _sample_generator():

        state_sequence = []
        state_sequence.append(state)

        env.state = state

        time = 0
        for t in range(time_horizon):
            action = policy(time=time, state=env.state)
            next_state, cost, done, _ = env.step(time=t, action=action)

            state_sequence.append(next_state)

            time += 1

        yield state_sequence

    return _sample_generator


def monte_carlo_sr(
    env: DynamicalSystem,
    policy: BasePolicy,
    T: np.ndarray,
    num_iterations: int = None,
    time_horizon: int = None,
    constraint_tube: list = None,
    target_tube: list = None,
    problem: str = "THT",
    verbose: bool = False,
):
    """Stochastic reachability using Monte-Carlo.

    Computes an approximation of the safety probabilities of the stochastic reachability
    problem using Monte-Carlo methods.

    Args:
        env: The dynamical system model.
        policy: The policy applied to the system during sampling.
        T: Points to estimate the safety probabilities at. Should be in the form of a
            2D-array, where each row indicates a point.
        num_iterations: Number of Monte-Carlo iterations.
        constraint_tube: List of spaces or constraint functions. Must be the same
            length as `num_steps`.
        target_tube: List of spaces or target functions. Must be the same length as
            `num_steps`.
        problem: One of `{"THT", "FHT"}`. `"THT"` specifies the terminal-hitting time
            problem and `"FHT"` specifies the first-hitting time problem.
        verbose: Boolean flag to indicate verbose output.

    """

    alg = MonteCarloSR(
        num_iterations=num_iterations,
        time_horizon=time_horizon,
        constraint_tube=constraint_tube,
        target_tube=target_tube,
        problem=problem,
        verbose=verbose,
    )
    return alg.fit_predict(env, policy, T)


class MonteCarloSR(RegressorMixin):
    """Stochastic reachability using Monte-Carlo.

    Computes an approximation of the safety probabilities of the stochastic reachability
    problem using Monte-Carlo methods.

    Args:
        num_iterations: Number of Monte-Carlo iterations.
        time_horizon: Time horizon of the algorithm.
        constraint_tube: List of spaces or constraint functions. Must be the same
            length as `num_steps`.
        target_tube: List of spaces or target functions. Must be the same length as
            `num_steps`.
        problem: One of `{"THT", "FHT"}`. `"THT"` specifies the terminal-hitting time
            problem and `"FHT"` specifies the first-hitting time problem.
        verbose: Boolean flag to indicate verbose output.

    """

    def __init__(
        self,
        num_iterations: int = None,
        time_horizon: int = None,
        constraint_tube: list = None,
        target_tube: list = None,
        problem: str = "THT",
        verbose: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.num_iterations = num_iterations
        self.time_horizon = time_horizon

        self.constraint_tube = constraint_tube
        self.target_tube = target_tube

        self.problem = problem

        self.verbose = verbose

    def _validate_params(self, S):

        assert self.num_iterations >= 0 and isinstance(
            self.num_iterations, (int, np.integer)
        )
        assert self.time_horizon >= 0 and isinstance(
            self.time_horizon, (int, np.integer)
        )

        if self.constraint_tube is None:
            raise ValueError("Must supply constraint tube.")

        if self.target_tube is None:
            raise ValueError("Must supply target tube.")

        if self.problem not in ("FHT", "THT"):
            raise ValueError(f"problem is not in {'THT', 'FHT'}")

        if self.problem == "THT":
            self.step_fn = _tht_step
        elif self.problem == "FHT":
            self.step_fn = _fht_step

    def _validate_data(self, S):

        if S is None:
            raise ValueError("Must supply a sample.")

    def fit(self):
        raise NotImplementedError

    def predict(self):
        raise NotImplementedError

    def fit_predict(self, env: DynamicalSystem, policy: BasePolicy, T: np.ndarray):
        """Run the algorithm.

        Computes the safety probabilities for the points provided. For each point in
        `T`, the algorithm computes a collection of trajectories using the point as the
        initial condition. Then, we can evaluate the indicator functions for each
        generated trajectory and the estimated safety probability is the sum of
        indicators divided by the number of trajectories.

        Args:
            env: The dynamical system model. Needed to configure the sampling spaces.
            T: Points to estimate the safety probabilities at. Should be in the form of
                a 2D-array, where each row indicates a point.

        Returns:
            The safety probabilities corresponding to each point. The output is in the
            form of a 2D-array, where each row corresponds to the points in `T` and the
            number of columns corresponds to the number of time steps.

        """

        self._validate_params(T)

        self._validate_data(T)
        num_test_points = len(T)
        T = np.array(T)

        pbar = ms_tqdm(
            total=num_test_points,
            bar_format=_progress_fmt,
            disable=False if self.verbose is True else True,
        )

        # Initialize the safety probability matrix.
        safety_probabilities = np.empty(
            (self.time_horizon, num_test_points), dtype=np.float32
        )

        for i, state in enumerate(T):

            # For each point, generate a collection of trajectories.
            S = sample(
                sampler=_monte_carlo_trajectory_sampler(
                    time_horizon=self.time_horizon, env=env, state=state, policy=policy
                ),
                sample_size=self.num_iterations,
            )

            # Working backwards in time, compute the "likelihood" that the trajectories
            # will satisfy the constraints set up by the constraint tube and target
            # tube.
            for t in range(self.time_horizon - 1, -1, -1):

                satisfies_constraints = _trajectory_indicator(
                    S,
                    time_horizon=t,
                    constraint_tube=self.constraint_tube,
                    target_tube=self.target_tube,
                    step_fn=self.step_fn,
                )

                safety_probabilities[t, i] = (
                    satisfies_constraints.sum() / self.num_iterations
                )

            pbar.update()

        pbar.close()

        # Return the flipped safety probabilities to be in line with kernel SR.
        return safety_probabilities[::-1, ...]



from abc import ABC, abstractmethod

import numpy as np


class ClassifierMixin(ABC):
    """Base class for algorithms.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from ClassifierMixin.

    The ClassifierMixin is meant to mimic the sklearn estimator base class in order
    to promote a standard interface among machine learning algorithms. It requires that
    all subclasses implement a ``fit``, ``predict``, and ``score`` method.

    """

    _estimator_type = "classifier"

    @abstractmethod
    def fit(self):
        raise NotImplementedError

    @abstractmethod
    def predict(self):
        raise NotImplementedError

    @abstractmethod
    def score(self):
        raise NotImplementedError


class ClusterMixin(ABC):
    """Base class for algorithms.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from ClusterMixin.

    The ClusterMixin is meant to mimic the sklearn estimator base class in order
    to promote a standard interface among machine learning algorithms. It requires that
    all subclasses implement a ``fit`` and ``fit_predict`` method.

    """

    _estimator_type = "clusterer"

    @abstractmethod
    def fit(self):
        raise NotImplementedError

    @abstractmethod
    def fit_predict(self):
        raise NotImplementedError


class RegressorMixin(ABC):
    """Base class for algorithms.

    This class is ABSTRACT, meaning it is not meant to be instantiated directly.
    Instead, define a new class that inherits from RegressorMixin.

    The RegressorMixin is meant to mimic the sklearn estimator base class in order
    to promote a standard interface among machine learning algorithms. It requires that
    all subclasses implement a ``fit``, ``predict``, and ``score`` method.

    """

    _estimator_type = "regressor"

    @abstractmethod
    def fit(self):
        raise NotImplementedError

    @abstractmethod
    def predict(self):
        raise NotImplementedError

    @abstractmethod
    def score(self):
        raise NotImplementedError



import gym

import numpy as np
from numpy.core.numeric import isscalar


def subspace(
    space: gym.spaces.Box, low, high, shape: tuple = None, seed: int = None
) -> gym.spaces.Box:

    # if shape is not None:
    #     shape = tuple(shape)

    #     # assert same as space or less than
    #     assert space.shape == shape, "shape must match space.shape"

    # elif not np.isscalar(low):
    #     shape = low.shape

    # elif not np.isscalar(high):
    #     shape = high.shape

    # if np.isscalar(low):
    #     low = np.full(shape, low, dtype=space.dtype)

    # if np.isscalar(high):
    #     high = np.full(shape, high, dtype=space.dtype)

    _space = gym.spaces.Box(low=low, high=high, shape=shape, dtype=space.dtype)

    if seed is not None:
        _space.seed(seed=seed)

    return _space
